"""Base class for particle pencil beam dose calculation algorithms."""

import warnings
from abc import abstractmethod
from typing import cast, Literal, Any
import logging
import time
from copy import deepcopy

import numpy as np
from scipy.interpolate import interp1d
from scipy.integrate import cumulative_trapezoid

from pyRadPlan.ct import CT
from pyRadPlan.stf import SteeringInformation
from pyRadPlan.machines import IonAccelerator, IonPencilBeamKernel, LateralCutOff
from pyRadPlan.cst import StructureSet
from ._base_pencilbeam import PencilBeamEngineAbstract


logger = logging.getLogger(__name__)


class ParticlePencilBeamEngineAbstract(PencilBeamEngineAbstract):
    """
    Abstract interface for Particle Pencil-Beam dose calculation.

    This class extends PencilBeamEngineAbstract by adding infrastructure for particles spots and
    quantities like LET and biological dose for variable RBE calculations.

    Attributes
    ----------
    calc_let : bool
        Boolean which defines if LET should be calculated.
    calc_bio_dose : bool
        Boolean to query biological dose calculation.
    air_offset_correction : bool
        Corrects WEPL for SSD difference to kernel database.
    lateral_model : Literal["auto", "single", "double", "multi", "fastest"]
        Lateral Model used. 'auto' uses the most accurate model available ( multiple Gaussians).
        "fastest" uses the most simple, inaccurate model (e.g., single Gaussian).
    cut_off_method : Literal["integral", "relative"]
        Method used to determine lateral cut off. 'integral' uses the integral of the lateral
        dose profile to obtain the cut-off distance. 'relative' finds the lateral distance at which
        the dose drops to the requested cutoff
    """

    calc_let: bool
    calc_bio_dose: bool
    air_offset_correction: bool
    lateral_model: Literal["auto", "single", "double", "multi", "fastest"]
    cut_off_method: Literal["integral", "relative"]

    def __init__(self, pln):
        self.calc_let = True
        self.calc_bio_dose = False
        self.air_offset_correction = True
        self.lateral_model = "fastest"
        self.cut_off_method = "integral"

        # Protected properties with public get access
        self._constant_rbe = None  # constant RBE value
        self._v_tissue_index = None  # Stores tissue indices available in the matRad base data
        self._v_alpha_x = None  # Stores Photon Alpha
        self._v_beta_x = None  # Stores Photon Beta

        super().__init__(pln)

    @abstractmethod
    def _calc_particle_bixel(self, bixel: dict):
        pass

    def _choose_lateral_model(self):
        """Choose & validate the lateral beam model."""

        available_models = {
            "single": self._machine.has_single_gaussian_kernel,
            "double": self._machine.has_double_gaussian_kernel,
            "multi": self._machine.has_multi_gaussian_kernel,
        }

        # First we validate the users choice
        if self.lateral_model in available_models:
            if not available_models[self.lateral_model]:
                logger.warning(
                    "Chosen Machine does not support a %s Gaussian Pencil-Beam model!",
                    self.lateral_model,
                )
                self.lateral_model = "auto"
        elif self.lateral_model not in ["auto", "fastest"]:
            logger.warning(
                "Invalid lateral model %s. Using auto instead.",
                self.lateral_model,
            )
            self.lateral_model = "auto"

        # Now if auto, we choose the most accurate one (i.e. the last one in the dictionary that
        # is available)

        if self.lateral_model == "auto":
            for key, available in available_models.items():
                if available:
                    self.lateral_model = key

        # if it is "auto"
        if self.lateral_model == "fastest":
            for key, available in available_models.items():
                if available:
                    self.lateral_model = key
                    break

        logger.info("Using a %s Gaussian pencil-beam kernel model!\n", self.lateral_model)

    def _compute_bixel(self, curr_ray: dict, k: int) -> dict:
        """
        Compute the bixel for the given ray and index.

        Parameters
        ----------
        curr_ray : dict
            The current ray data.
        k : int
            The index of the bixel.

        Returns
        -------
        dict
            The computed bixel.
        """
        # Initialize Bixel Geometry
        bixel = self._init_bixel(curr_ray, k)

        # Compute Bixel
        self._calc_particle_bixel(bixel)

        return bixel

    def _init_bixel(self, curr_ray, k):
        """
        Initialize general bixel geometry for particle dose calculation.

        Parameters
        ----------
        curr_ray : dict
            The current ray data.
        k : int
            The index of the bixel.

        Returns
        -------
        dict
            The initialized bixel.
        """

        bixel = curr_ray["beamlets"][k]
        bixel["beam_index"] = curr_ray["beam_index"]
        bixel["ray_index"] = curr_ray["ray_index"]
        bixel["bixel_index"] = k

        # First we get metadata: MU, corresponding base data entry, etc.
        # Extract MU data if present (checks for downwards compatibility)
        bixel["min_mu"] = 0
        if "min_mu" in curr_ray:
            bixel["min_mu"] = curr_ray["min_mu"][k]

        bixel["max_mu"] = float("inf")
        if "max_mu" in curr_ray:
            bixel["max_mu"] = curr_ray["max_mu"][k]

        bixel["num_particles_per_mu"] = 1e6
        if "num_particles_per_mu" in curr_ray:
            bixel["num_particles_per_mu"] = curr_ray["num_particles_per_mu"][k]

        # Find energy index in base data
        energy = curr_ray["beamlets"][k]["energy"]
        energy_ix = self._machine.get_energy_index(energy, 4)

        if energy_ix.size > 1:
            raise ValueError("Multiple energies found in base data for one bixel!")

        energy_ix = np.int64(energy_ix)

        bixel["energy_ix"] = energy_ix

        # Get the kernel for the current energy
        tmp_machine = cast(IonAccelerator, self._machine)
        bixel["kernel"] = tmp_machine.get_kernel_by_index(energy_ix)

        bixel["range_shifter"] = curr_ray["beamlets"][k]["range_shifter"]
        bixel["SSD"] = curr_ray["SSD"]
        bixel["rad_depth_offset"] = curr_ray["rad_depth_offset"]

        # Compute initial spot width
        bixel["sigma_ini_sq"] = curr_ray["sigma_ini"][k] ** 2

        # Apply beam modifiers
        self._get_beam_modifiers(bixel)

        # Gets bixel.ix (voxel indices) and bixel.subIx (logical
        # indices to be used) after cutoff. Storing these allows us to
        # use indexing for performance and avoid too many copies
        self._get_bixel_indices_on_ray(bixel, curr_ray)
        if "ix" not in bixel or bixel["ix"].size == 0:
            return bixel

        # Get quantities 1:1 from ray. Here we trust Python's memory
        # management to not copy the arrays until they are modified.
        # This allows us to efficiently access them by indexing in the
        # bixel computation
        bixel["radial_dist_sq"] = curr_ray["radial_dist_sq"][bixel["sub_ix"]]
        bixel["rad_depths"] = curr_ray["rad_depths"][bixel["sub_ix"]]
        # TODO:
        # if self.calc_bio_dose:
        #     bixel["v_tissue_index"] = curr_ray["v_tissue_index"][bixel["sub_ix"]]
        #     bixel["v_alpha_x"] = curr_ray["v_alpha_x"][bixel["sub_ix"]]
        #     bixel["v_beta_x"] = curr_ray["v_beta_x"][bixel["sub_ix"]]

        return bixel

    def _interpolate_kernels_in_depth(self, bixel):
        kernel = cast(IonPencilBeamKernel, bixel["kernel"])
        depths = kernel.depths

        # Add potential offset
        depths = depths + kernel.offset - bixel["rad_depth_offset"]

        # Conversion factor from MeV cm^2/g per primary to Gy mm^2 per 1e6 primaries
        conversion_factor = 1.6021766208e-02

        # Find all values we need to interpolate
        used_kernels = {}
        used_kernels["idd"] = conversion_factor * kernel.idd

        # Lateral Kernel Model
        if self.lateral_model == "single":
            used_kernels["sigma"] = kernel.sigma
        elif self.lateral_model == "double":
            used_kernels["sigma_1"] = kernel.sigma_1
            used_kernels["sigma_2"] = kernel.sigma_2
            used_kernels["weight"] = kernel.weight
        elif self.lateral_model == "multi":
            used_kernels["weight_multi"] = kernel.weight_multi
            used_kernels["sigma_multi"] = kernel.sigma_multi
        else:
            raise ValueError("Invalid Lateral Model")

        # LET
        if self.calc_let:
            used_kernels["let"] = kernel.let

        # bioDose
        # TODO:
        if self.calc_bio_dose:
            used_kernels["alpha"] = kernel.alpha
            used_kernels["beta"] = kernel.beta

        # Interpolate all fields in X
        kernel_interp = {}
        for key, kernel_data in used_kernels.items():
            if kernel_data.ndim > 1:
                tmp_kernel_data = np.apply_along_axis(
                    lambda x: np.interp(bixel["rad_depths"], depths, x), axis=1, arr=kernel_data
                )
            else:
                tmp_kernel_data = np.interp(bixel["rad_depths"], depths, kernel_data)

            kernel_interp[key] = tmp_kernel_data

        return kernel_interp

    def _get_ray_geometry_from_beam(self, ray: dict[str], beam_info: dict[str]):
        lateral_ray_cutoff = self._get_lateral_distance_from_dose_cutoff_on_ray(ray)

        # Ray tracing for beam i and ray j
        ix, radial_dist_sq, _, _ = self.calc_geo_dists(
            beam_info["bev_coords"],
            ray["source_point_bev"],
            ray["target_point_bev"],
            ray["sad"],
            beam_info["valid_coords_all"],
            lateral_ray_cutoff,
        )

        # Subindex given the relevant indices from the geometric distance calculation
        ray["valid_coords"] = [beam_ix & ix for beam_ix in beam_info["valid_coords"]]
        ray["ix"] = [self._vdose_grid[ix_in_grid] for ix_in_grid in ray["valid_coords"]]

        ray["radial_dist_sq"] = [
            radial_dist_sq[beam_ix[ix]] for beam_ix in beam_info["valid_coords"]
        ]

        ray["valid_coords_all"] = np.any(np.vstack(ray["valid_coords"]), axis=1)

        ray["geo_depths"] = [
            rD[ix] for rD, ix in zip(beam_info["geo_depths"], ray["valid_coords"])
        ]  # usually not needed for particle beams
        ray["rad_depths"] = [
            rD[ix] for rD, ix in zip(beam_info["rad_depths"], ray["valid_coords"])
        ]

    def _get_bixel_indices_on_ray(self, curr_bixel, curr_ray):
        kernel = cast(IonPencilBeamKernel, curr_bixel["kernel"])

        # Create offset vector to account for additional offsets modeled in the base data
        # and a potential range shifter
        tmp_offset = kernel.offset - curr_bixel["rad_depth_offset"]

        # Find depth-dependent lateral cutoff
        if self.dosimetric_lateral_cutoff == 1:
            curr_ix = curr_ray["rad_depths"] <= kernel.depths[-1] + tmp_offset
        elif 0 < self.dosimetric_lateral_cutoff < 1:
            cutoff_info = kernel.lateral_cut_off

            if cutoff_info.cut_off.size > 1:
                curr_ix = (
                    np.interp(
                        curr_ray["rad_depths"],
                        cutoff_info.depths + tmp_offset,
                        cutoff_info.cut_off**2,
                        left=np.nan,
                        right=np.nan,
                    )
                    >= curr_ray["radial_dist_sq"]
                ) & (curr_ray["rad_depths"] <= kernel.depths[-1] + tmp_offset)
            else:
                curr_ix = (cutoff_info.cut_off[0] ** 2 >= curr_ray["radial_dist_sq"]) & (
                    curr_ray["rad_depths"] <= kernel.depths[-1] + tmp_offset
                )
        else:
            raise ValueError("dosimetric_lateral_cutoff must be a value > 0 and <= 1!")

        curr_bixel["sub_ix"] = curr_ix
        curr_bixel["ix"] = curr_ray["ix"][curr_ix]

    def _get_beam_modifiers(self, curr_bixel):
        add_sigma_sq = 0
        rad_depth_offset = 0

        # Consider range shifter for protons if applicable
        if curr_bixel["range_shifter"]["eq_thickness"] > 0:
            # TODO: We should do this check in dose calc initialization instead to spare some time
            if self._machine["meta"]["radiationMode"] != "protons":
                warnings.warn(
                    "Range shifter not valid for irradiation with particle other than protons!"
                )

            # Compute!
            sigma_rashi = self._calc_sigma_rashi(
                curr_bixel["kernel"]["energy"], curr_bixel["range_shifter"], curr_bixel["SSD"]
            )

            # Add to initial sigma in quadrature
            add_sigma_sq += sigma_rashi**2
            rad_depth_offset += curr_bixel["range_shifter"]["eq_thickness"]

        curr_bixel["add_sigma_sq"] = add_sigma_sq
        curr_bixel["rad_depth_offset"] += rad_depth_offset

    def _init_dose_calc(self, ct, cst, stf):
        """
        Initialize dose calculation.

        Modified inherited method of the superclass DoseEngine,
        containing intialization which are specificly needed for
        pencil beam calculation and not for other engines.
        """

        dij = super()._init_dose_calc(ct, cst, stf)

        # Choose the lateral pencil beam model
        self._choose_lateral_model()

        # Toggles correction of small difference of current SSD to distance used
        # in generation of base data (e.g. phantom surface at isocenter)

        # TODO: this is a dummy. bio_param not implemented yet...
        self.bio_param = {"bioOpt": False}

        # Omit field checks of fit_air_offset and BAMStoIsoDist as validated through machine model

        # biology
        if hasattr(self, "_constant_rbe") and self._constant_rbe is not None:
            dij["RBE"] = self._constant_rbe

        # TODO: (Comment from matlab): This is clumsy and needs to be changed with the biomodel
        # update
        if self.bio_param["bioOpt"]:
            self.calc_bio_dose = True

        # Load biologicla base data if needed
        if self.calc_bio_dose:
            dij = self._load_biological_kernel(cst, dij)  # TODO: Not fully implemented yet
            # allocate alpha and beta dose container and sparse matrices in the dij struct,
            # for more informations see corresponding method
            dij = self._allocate_bio_dose_container(dij)  # TODO: Not fully implemented yet

        # Allocate LET containner and let sparse matrix in dij struct
        if self.calc_let:
            if self._machine.has_let_kernel:
                dij = self._allocate_let_container(dij)
            else:
                logger.warning(
                    "No LET data found in machine data. LET calculation will be skipped."
                )
                self.calc_let = False

        self._effective_lateral_cutoff = self.geometric_lateral_cutoff

        return dij

    def _init_beam(
        self, dij: dict, ct: CT, cst: StructureSet, stf: SteeringInformation, i
    ) -> dict:
        """
        Initialize beam for dose calculation.

        Extends the inherited method with particle-specific initialization.

        Parameters
        ----------
        dij : dict
            The dose influence matrix dictionary.
        ct : CT
            The CT object.
        _cst : StructureSet
            The structure set object. Unused here
        stf : SteeringInformation
            The steering information object.
        i : int
            Index of the beam.

        Returns
        -------
        dict
            Updated Beam Information dictionary.
        """
        beam_info = super()._init_beam(dij, ct, cst, stf, i)

        # Sanity Check
        assert isinstance(self._machine, IonAccelerator)

        # Assuming currBeam is part of beam_info
        curr_beam = beam_info["beam"]

        # Find max energy
        max_energy = max(
            beamlet["energy"] for ray in curr_beam["rays"] for beamlet in ray["beamlets"]
        )

        max_energy_ix = self._machine.get_energy_index(max_energy, 4)

        max_energy_kernel = self._machine.get_kernel_by_index(max_energy_ix)

        # Find minimum range shifter equivalent thickness
        min_ra_shi = min(
            beamlet["range_shifter"]["eq_thickness"]
            for ray in curr_beam["rays"]
            for beamlet in ray["beamlets"]
        )
        rad_depth_offset = max_energy_kernel.offset + min_ra_shi

        # Apply limit in depth
        sub_select_ix = [
            rD < (max_energy_kernel.depths[-1] - rad_depth_offset)
            for rD in beam_info["rad_depths"]
        ]

        for scen in range(len(beam_info["valid_coords"])):
            beam_info["valid_coords"][scen] = np.array(
                [
                    True if ss & vc else False
                    for ss, vc in zip(sub_select_ix[scen], beam_info["valid_coords"][scen])
                ]
            )

        beam_info["valid_coords_all"] = np.any(np.column_stack(beam_info["valid_coords"]), axis=1)

        # Precompute CutOff
        logger.info("Calculating lateral cutoffs for beam %d...", i)
        t_start = time.time()
        self._calc_lateral_particle_cut_off(self.dosimetric_lateral_cutoff, beam_info)
        t_elapsed = time.time() - t_start
        logger.info("Done in %f seconds.", t_elapsed)

        return beam_info

    def _load_biological_kernel(self, _cst: StructureSet, dij: dict[str, Any]):
        """
        Load / organize biological kernel data from the machine.

        Stores information in the dose influence matrix dictionary.

        Parameters
        ----------
        _cst : StructureSet
            The structure set object.
        dij : dict
            The dose influence matrix dictionary.

        Returns
        -------
        dict
            The updated dose influence matrix dictionary with biological information.
        """
        logger.warning(
            "Biological Kernel not implemented yet. "
            "Biological dose calculation will not honor tissue settings."
        )
        return dij

    def _allocate_bio_dose_container(self, dij: dict[str, Any]):
        """
        Allocate space for container used in LET calculation.

        Parameters
        ----------
        dij : dict
            The dose influence matrix dictionary.

        Returns
        -------
        dict
            The updated dose influence matrix dictionary with LET containers allocated.
        """

        if self._machine.has_alpha_beta_kernels:
            dij = self._allocate_quantity_matrices(dij, ["alpha_dose", "sqrt_beta_dose"])
        else:
            warnings.warn(
                "Biological kernels not available. Biological dose will not be calculated."
            )

        return dij

    def _allocate_let_container(self, dij):
        """
        Allocate space for container used in LET calculation.

        Parameters
        ----------
        dij : dict
            The dose influence matrix dictionary.

        Returns
        -------
        dict
            The updated dose influence matrix dictionary with LET containers allocated.
        """
        # Get MatRad Config instance for displaying warnings

        if self._machine.has_let_kernel:
            dij = self._allocate_quantity_matrices(dij, ["let_dose"])
        else:
            warnings.warn("LET not available in the machine data. LET will not be calculated.")

        return dij

    def _calc_lateral_particle_cut_off(self, cut_off_level, stf_element):
        # Sanity Checks
        assert isinstance(self._machine, IonAccelerator)

        if len(stf_element) > 1 and not isinstance(stf_element, dict):
            raise ValueError(
                "CutOff can only be precalculated for a single element, "
                "but you provided steering information for multiple beams!"
            )

        if cut_off_level <= 0.98:
            logger.warning(
                "A lateral cut off below 0.98 may result in an inaccurate dose calculation"
            )

        if cut_off_level < 0 or cut_off_level > 1:
            warnings.warn("Lateral cutoff is out of range - using default cut off of 0.99")
            cut_off_level = 0.99

        # conversion factor from kernel units (MeV cm^2/g per primary) to Gy mm^2 per 1e6 primaries
        # TODO: this factor should probably be in the machine data and translate to MU
        conversion_factor = 1.6021766208e-02

        # Define some variables needed for the cutoff calculation
        v_x = np.concatenate(([0], np.logspace(-1, 3, 1200)))  # [mm]

        # Integration steps
        r_mid = 0.5 * (v_x[:-1] + v_x[1:])  # [mm]
        dr = np.diff(v_x)
        radial_dist_sq = r_mid**2

        # Number of depth points for which a lateral cutoff is determined
        num_depth_val = 35

        # Extract SSD for each bixel
        v_ssd = np.ones(
            len(
                [
                    beamlet["energy"]
                    for ray in stf_element["beam"]["rays"]
                    for beamlet in ray["beamlets"]
                ]
            )
        )

        cnt = 0
        # TODO: this can be done more efficiently. (instead of calc len() all the time)
        # TODO: these for loops can be simplified!
        for ray in stf_element["beam"]["rays"]:
            len_ray_energy = len([beamlet["energy"] for beamlet in ray["beamlets"]])
            v_ssd[cnt : cnt + len_ray_energy] = ray["SSD"]
            # TODO: whats the purpose of this? in matRad it counts the len(v_ssd)+1 for some reason
            cnt += len_ray_energy

        # Setup energy, focus index, sigma look up table - only consider unique rows
        energy_sigma_lut, ix_unique = np.unique(
            np.column_stack(
                (
                    [
                        beamlet["energy"]
                        for ray in stf_element["beam"]["rays"]
                        for beamlet in ray["beamlets"]
                    ],
                    [
                        beamlet["focus_ix"]
                        for ray in stf_element["beam"]["rays"]
                        for beamlet in ray["beamlets"]
                    ],
                    v_ssd,
                )
            ),
            axis=0,
            return_index=True,
        )
        # add extra dimension for energy_sigma_lut
        # TODO: there must be a simpler way to do this
        energy_sigma_lut = np.hstack(
            (energy_sigma_lut, np.full((len(energy_sigma_lut), 1), np.nan))
        )

        range_shifter_lut = [
            beamlet["range_shifter"]
            for ray in stf_element["beam"]["rays"]
            for beamlet in ray["beamlets"]
        ]
        range_shifter_lut = [range_shifter_lut[i] for i in ix_unique]

        # Find the largest initial beam width considering focus index, SSD and range shifter
        # for each individual energy
        for i in range(len(energy_sigma_lut)):
            # find index of maximum used energy (round to keV for numerical reasons)
            energy_ix = self._machine.get_energy_index(energy_sigma_lut[i, 0], 4)
            # Get the kernel entry
            kernel = self._machine.get_kernel_by_index(energy_ix)
            # Get the available beam foci
            foci = self._machine.get_foci_by_index(energy_ix)

            focus_ix = int(energy_sigma_lut[i, 1])
            focus = foci[focus_ix]

            sigma_ini = np.interp(energy_sigma_lut[i, 2], focus.dist, focus.sigma)

            sigma_ini_sq = sigma_ini**2

            # Consider range shifter for protons if applicable
            if (
                self._machine.radiation_mode == "protons"
                and range_shifter_lut[i]["eq_thickness"] > 0
            ):
                # TODO: this is not tested yet as only Generic is used
                sigma_rashi = self._calc_sigma_rashi(
                    self._machine.energies[energy_ix],
                    range_shifter_lut[i],
                    energy_sigma_lut[i, 2],
                )
                sigma_ini_sq += sigma_rashi**2

            energy_sigma_lut[i, 3] = sigma_ini_sq

        # Find for each individual energy the broadest initial beam width
        unique_energies = np.unique(energy_sigma_lut[:, 0])
        largest_sigma_sq4unique_energies = np.full(len(unique_energies), np.nan)
        ix_max = np.full(len(unique_energies), np.nan)
        for i, energy in enumerate(unique_energies):
            largest_sigma_sq4unique_energies[i] = np.max(
                energy_sigma_lut[energy == energy_sigma_lut[:, 0], 3]
            )
            ix_max[i] = np.argmax(energy_sigma_lut[energy == energy_sigma_lut[:, 0], 3])

        # Get energy indices for looping
        v_energies_ix = np.where(np.isin(self._machine.energies, unique_energies))[0]
        cnt = 0

        # Loop over all entries in the machine.data struct
        for energy_ix in v_energies_ix:
            # Get the kernel entry
            kernel = self._machine.get_kernel_by_index(energy_ix)

            # set default depth cut off - finite value will be set during first iteration
            depth_dose_cut_off = np.inf

            # Get the current integrated depth dose profile
            idd_org = conversion_factor * kernel.idd
            peak_ix_org = np.argmax(idd_org)

            # Get indices for which a lateral cutoff should be calculated
            cum_int_energy = cumulative_trapezoid(idd_org, kernel.depths, initial=0)

            peak_tail_relation = 0.5
            num_depth_val_to_peak = int(np.ceil(num_depth_val * peak_tail_relation))
            num_depth_val_tail = int(np.ceil(num_depth_val * (1 - peak_tail_relation)))
            energy_steps_to_peak = cum_int_energy[peak_ix_org] / num_depth_val_to_peak
            energy_steps_tail = (
                cum_int_energy[-1] - cum_int_energy[peak_ix_org]
            ) / num_depth_val_tail

            v_energy_steps = np.unique(
                np.concatenate(
                    (
                        np.arange(0, cum_int_energy[peak_ix_org], energy_steps_to_peak),
                        [cum_int_energy[peak_ix_org]],
                        np.arange(
                            cum_int_energy[peak_ix_org + 1], cum_int_energy[-1], energy_steps_tail
                        ),
                        [cum_int_energy[-1]],
                    )
                )
            )

            cum_int_energy, ix = np.unique(cum_int_energy, return_index=True)

            depth_values = np.interp(v_energy_steps, cum_int_energy, kernel.depths[ix])
            idd = conversion_factor * np.interp(depth_values, kernel.depths, kernel.idd)

            cnt += 1

            # Create a default cutoff and store it in the kernel
            kernel.lateral_cut_off = LateralCutOff(
                comp_fac=1.0, depths=depth_values, cut_off=np.inf * np.ones_like(depth_values)
            )

            base_kernel = deepcopy(kernel)  # similar to base_data in matRad

            # TODO: this could probably be done for all depths with one calculation without a loop
            for j, current_depth in enumerate(depth_values):
                # If there's no cut-off set, we do not need to find it. Set it to infinity
                if cut_off_level == 1:
                    continue
                # Create a dummy bixel calculating radial dose distribution for the current
                # depth
                bixel = {
                    "energy_ix": energy_ix,
                    "kernel": base_kernel,
                    "radial_dist_sq": radial_dist_sq,
                    "sigma_ini_sq": largest_sigma_sq4unique_energies[
                        cnt - 1
                    ],  # TODO: check if this result is correct (rounded but may be right)
                    "rad_depths": (current_depth + base_kernel.offset)
                    * np.ones_like(radial_dist_sq),
                    "v_tissue_index": np.ones_like(radial_dist_sq),
                    "v_alpha_x": 0.5 * np.ones_like(radial_dist_sq),
                    "v_beta_x": 0.05 * np.ones_like(radial_dist_sq),
                    "sub_ray_ix": np.ones_like(radial_dist_sq, dtype=bool),
                    "ix": np.arange(len(radial_dist_sq)),
                    "rad_depth_offset": 0,
                    "add_sigma_sq": 0,
                }

                self._calc_particle_bixel(bixel)
                dose_r = bixel["physical_dose"]

                # Do an integration check that the radial dose integrates to the tabulated
                # integrated depth dose
                cum_area = np.cumsum(2 * np.pi * r_mid * dose_r * dr)
                relative_tolerance = 0.5  # in [%]

                if abs((cum_area[-1] / idd[j]) - 1) * 100 > relative_tolerance:
                    warnings.warn("Shell integration in cut-off calculation is inconsistent!")

                # obtain the cut-off compensation factor dending on the cut-off method
                if self.cut_off_method == "integral":
                    ix = (cum_area >= idd[j] * cut_off_level).argmax()
                    kernel.lateral_cut_off.comp_fac = cut_off_level**-1
                elif self.cut_off_method == "relative":
                    ix = (dose_r <= (1 - cut_off_level) * np.max(dose_r)).argmax()
                    rel_fac = cum_area[ix] / cum_area[-1]
                    kernel.lateral_cut_off.comp_fac = rel_fac**-1
                else:
                    raise ValueError("Invalid Cutoff Method. Must be 'integral' or 'relative'!")
                # Nothing was cut-off -> warn the user
                if ix == len(cum_area):
                    depth_dose_cut_off = np.inf
                    warnings.warn("Couldn't find lateral cut off!")
                else:
                    depth_dose_cut_off = r_mid[ix]

                kernel.lateral_cut_off.cut_off[j] = depth_dose_cut_off

            self._machine.update_kernel_at_index(energy_ix, kernel)

    def _init_ray(self, beam_info: dict[str], j: int) -> dict[str]:
        ray = super()._init_ray(beam_info, j)

        self._machine = cast(IonAccelerator, self._machine)

        # Calculate initial sigma for all bixels on the current ray
        # TODO: here [ray] since calc_sigma_ini takes multiple rays (why?)
        ray["sigma_ini"] = self._calc_sigma_ini_on_ray(ray)

        # Since pyRadPlan's ray cast starts at the skin and base data
        # is generated at some source to phantom distance
        # we can explicitly correct for the nozzle to air WEPL in
        # the current case.
        if self.air_offset_correction:
            nozzle_to_skin = (ray["SSD"] + self._machine.bams_to_iso_dist) - self._machine.sad
            ray["rad_depth_offset"] = 0.0011 * (nozzle_to_skin - self._machine.fit_air_offset)
        else:
            ray["rad_depth_offset"] = 0

        # Just use tissue classes of voxels found by ray tracer
        # TODO: not tested
        # if self.calc_bio_dose:
        #     for s in range(len(self._v_tissue_index)):
        #         ray["vtissue_index"][s] = self._v_tissue_index[s][ray["valid_coords"][s], :]
        #         ray["valpha_x"][s] = self._v_alpha_x[s][ray["valid_coords"][s]]
        #         ray["vbeta_x"][s] = self._v_beta_x[s][ray["valid_coords"][s]]

        return ray

    def _extract_single_scenario_ray(self, ray, scen_idx):
        scen_ray = super()._extract_single_scenario_ray(ray, scen_idx)
        # TODO: Add multscen support
        # Gets number of scenario
        scen_num = 1  # self.mult_scen['scenNum'][scen_idx]
        ct_scen = self.mult_scen.linear_mask[0][scen_num]

        if "vTissueIndex" in scen_ray:
            scen_ray["vTissueIndex"] = scen_ray["vTissueIndex"][ct_scen]
            scen_ray["vAlphaX"] = scen_ray["vAlphaX"][ct_scen]
            scen_ray["vBetaX"] = scen_ray["vBetaX"][ct_scen]

        return scen_ray

    def _fill_dij(
        self, bixel, dij, stf, scen_idx, curr_beam_idx, curr_ray_idx, curr_bixel_idx, bixel_counter
    ):
        super()._fill_dij(
            bixel, dij, stf, scen_idx, curr_beam_idx, curr_ray_idx, curr_bixel_idx, bixel_counter
        )

        # Add MU information
        if not self._calc_dose_direct:
            dij["min_mu"][bixel_counter] = bixel["min_mu"]
            dij["max_mu"][bixel_counter] = bixel["max_mu"]
            dij["num_of_particles_per_mu"][bixel_counter] = bixel["num_particles_per_mu"]

    def _get_lateral_distance_from_dose_cutoff_on_ray(self, ray: dict):
        # Find index of maximum used energy (round to keV for numerical reasons)
        self._machine = cast(IonAccelerator, self._machine)

        max_energy = max([beamlet["energy"] for beamlet in ray["beamlets"]])
        kernel = self._machine.get_kernel_by_energy(max_energy)
        lat_cut_off = kernel.lateral_cut_off

        # Get the lateral cutoff distance
        lateral_ray_cut_off = lat_cut_off.cut_off.max()

        return lateral_ray_cut_off

    def _calc_sigma_ini_on_ray(self, ray: dict, ssd: np.floating = None):
        """
        Get initial beam width (sigma) at the sruface.

        Parameters
        ----------
        ray : dict
            The ray data.
        ssd : float
            The source to surface distance. Can be omitted if ray has SSD stored

        Returns
        -------
        np.ndarray
            The initial sigma for the ray energies
        """

        if ssd is None:
            if "SSD" not in ray:
                raise ValueError("SSD not provided and not stored in ray data!")
            ssd = ray["SSD"]

        # energies on ray
        sigma_ini = np.zeros((len(ray["beamlets"]),), dtype=np.float64)
        for i, beamlet in enumerate(ray["beamlets"]):
            # obtain focus
            energy = beamlet["energy"]
            focus_ix = beamlet["focus_ix"]
            sigma_ini[i] = self._calc_sigma_ini_focus_ix(energy, focus_ix, ssd)
        return sigma_ini

    def _calc_sigma_ini_focus_ix(
        self, energy: np.float64, focus_ix: int, ssd: np.floating
    ) -> np.float64:
        """
        Get the initial beam width for a specific focus.

        Parameters
        ----------
        energy : np.float64
            The energy of the beam.
        focus_ix : int
            The focus index.
        ssd : np.floating
            The source to surface distance.

        Returns
        -------
        np.float64
            The initial sigma for the given energy and focus index.
        """
        focus = self._machine.foci[energy][focus_ix]
        interpolator = interp1d(focus.dist, focus.sigma, fill_value="extrapolate")
        return interpolator(ssd)

    def _calc_sigma_rashi(self, bd_entry, range_shifter: dict, ssd: float):
        # Distance of range shifter to patient surface
        rashi_dist = ssd - range_shifter["sourceRashiDistance"]

        # Convert to cm
        z_rs_cm = rashi_dist / 10.0  # [cm]
        t = range_shifter["eqThickness"] / 10.0  # [cm]

        # Constants
        a1 = 0.21  # [1]
        a2 = 0.77  # [1]
        c0 = 0.0191027  # [cm]
        c1 = 0.0204539  # [1]
        alpha = 0.0022  # [cm MeV ^ (-p)]
        p = 1.77  # [1]

        # Determine range
        if "range" in bd_entry:
            range_in_water = bd_entry["range"]
        else:
            energy = bd_entry["energy"]
            range_in_water = alpha * (energy**p)

        # Check if valid computation possible or range shifter too thick
        if t / range_in_water >= 0.95:
            raise ValueError(
                "Computation of range shifter sigma invalid. Range shifter is too thick."
            )

        # Improved HONG's Formula [1, Eq. 21]
        s = t / range_in_water
        sigma_t = (a1 * s + a2 * s**2) * (c0 + c1 * range_in_water)

        # Constants for further calculations
        c1 = 13.6  # MeV
        c2 = 0.038
        lr_water = 36.08  # rad. length [cm]
        lr_pmma = 40.55  # rad. length [cm]
        rsp_pmma = 1.165  # rSP [rel. u.]
        prsf_pmma = np.sqrt(
            rsp_pmma**3 * lr_water / lr_pmma * (1 + 2 * c2 * np.log(rsp_pmma * lr_water / lr_pmma))
        )

        # Calculate F1 and F2
        f1_part1 = (2 * c1**2 * alpha ** (2 / p)) / (4 * lr_pmma * (2 / p - 1))
        f1_part2 = ((1 - s) ** (2 - 2 / p) - 1) / (2 / p - 2) - s
        f1_part3 = range_in_water ** (2 - 2 / p)
        f1 = f1_part1 * f1_part2 * f1_part3

        f2_part1 = (c1**2 * alpha ** (2 / p)) / (4 * lr_pmma * (2 / p - 1))
        f2_part2 = ((1 - s) ** (1 - 2 / p) - 1) / 1
        f2_part3 = range_in_water ** (1 - 2 / p)
        f2 = f2_part1 * f2_part2 * f2_part3

        sigma_proj_sq = prsf_pmma**2 * (
            sigma_t**2 + f1 * z_rs_cm * rsp_pmma + f2 * (z_rs_cm * rsp_pmma) ** 2
        )  # [cm ^ 2]

        # Convert to mm
        sigma_rashi = 10.0 * np.sqrt(sigma_proj_sq)  # [mm]

        return sigma_rashi

    def round2(self, a: np.floating, b: int) -> np.floating:
        """
        Round a number stably for energy selection (helper function).

        Parameters
        ----------
        a : np.floating
            The number to round.
        b : int
            The number of decimal places to round to.

        Returns
        -------
        np.floating
            The rounded number.
        """
        return np.round(a * 10**b) / 10**b



import os
from abc import ABC
from abc import abstractmethod


class MatRadEngine(ABC):
    @property
    def matRadPath(self):
        return self._matRadPath

    @matRadPath.setter
    def matRadPath(self, path):
        # Check if valid matrad path
        isdir = os.path.isdir(path)
        if not isdir:
            raise ValueError("matRad not found")

        self._matRadPath = path

    @property
    @abstractmethod
    def matRadVersionStr(self):
        pass

    @property
    @abstractmethod
    def engine(self):
        pass

    @property
    @abstractmethod
    def engineVersionStr(self):
        pass

    def __getattr__(self, name):
        return getattr(self.engine, name)



from typing import Union, cast
import time
import logging

import numpy as np
from numpy.typing import NDArray


from ...plan import Plan
from ._optiprob import NonLinearPlanningProblem
from ..solvers import NonLinearOptimizer
from ..objectives import Objective

logger = logging.getLogger(__name__)


class NonLinearFluencePlanningProblem(NonLinearPlanningProblem):
    """
    Non-linear fluence-based planning problem.

    Parameters
    ----------
    pln : Union[Plan, dict], optional
        Plan object or dictionary to initialize the problem with.

    Attributes
    ----------
    bypass_objective_jacobian : bool, optional, default=True
        Whether to bypass the objective jacobian calculation. This is usefull for scalarized
        optimization (e.g. weighted sum of objectives) as it will minimize storage to only a single
        gradient vector per quantity.
    """

    name = "Non-Linear Fluence Planning Problem"
    short_name = "nonlin_fluence"

    bypass_objective_jacobian: bool

    def __init__(self, pln: Union[Plan, dict] = None):
        self.bypass_objective_jacobian = True

        self._target_voxels = None
        self._patient_voxels = None

        self._grad_cache_intermediate = None
        self._grad_cache = None
        self._obj_times = []
        self._deriv_times = []
        self._solve_time = None

        super().__init__(pln)

    def _initialize(self):
        """Initialize this problem."""
        super()._initialize()

        # Check if the solver is adequate to solve this problem
        # TODO: check that it can do constraints
        if not isinstance(self.solver, NonLinearOptimizer):
            raise ValueError("Solver must be an instance of SolverBase")

        self.solver.objective = self._objective_function
        self.solver.gradient = self._objective_gradient
        self.solver.bounds = (0.0, np.inf)
        self.solver.max_iter = 500

    def _objective_functions(self, x: NDArray) -> NDArray:
        """Define the objective functions."""

        q_vectors = {}
        q_scenarios = {}

        # Check & get Caches
        for q in self._quantities:
            q_vectors[q.identifier] = q(x)
            q_scenarios[q.identifier] = q.scenarios

        # Loop over all objectives
        f_vals = []

        for obj_info in self._objective_list:
            ix = obj_info[0]
            tmp_obj_list = cast(list[Objective], obj_info[1])
            for obj in tmp_obj_list:
                f_vals.append(
                    sum(
                        [
                            obj.priority
                            * obj.compute_objective(q_vectors[obj.quantity].flat[scen_ix][ix])
                            for scen_ix in q_scenarios[obj.quantity]
                        ]
                    )
                )

        # return as numpy array
        return np.asarray(f_vals, dtype=np.float64)

    def _objective_function(self, x: NDArray) -> np.float64:
        t = time.time()
        f = np.sum(self._objective_functions(x))
        self._obj_times.append(time.time() - t)
        return f

    def _objective_jacobian(self, x: NDArray) -> NDArray:
        """Define the objective jacobian."""

        q_vectors = {}
        q_scenarios = {}
        if self._grad_cache_intermediate is None:
            initialize_cache = True
            self._grad_cache_intermediate = {}
        else:
            initialize_cache = False

        # Check & get Caches
        for q in self._quantities:
            q_vectors[q.identifier] = q(x)
            q_scenarios[q.identifier] = q.scenarios
            if initialize_cache:
                if self.bypass_objective_jacobian:
                    cache_rows = 1
                else:
                    cache_rows = len(self._objectives_per_quantity[q.identifier])
                self._grad_cache_intermediate[q.identifier] = np.zeros(
                    (
                        cache_rows,
                        self._dij.dose_grid.num_voxels,
                    ),
                    dtype=np.float32,
                )
            else:
                self._grad_cache_intermediate[q.identifier].fill(0.0)

        cnt = 0
        for obj_info in self._objective_list:
            ix = obj_info[0]
            tmp_obj_list = cast(list[Objective], obj_info[1])
            for obj in tmp_obj_list:
                if self.bypass_objective_jacobian:
                    q_cache_index = 0
                else:
                    q_cache_index = self._q_cache_index[cnt]
                for scen_ix in q_scenarios[obj.quantity]:
                    self._grad_cache_intermediate[obj.quantity][q_cache_index, ix] += (
                        obj.priority
                        * obj.compute_gradient(q_vectors[obj.quantity].flat[scen_ix][ix])
                    )
                cnt += 1

        # perform chain rule and store in cache
        if self._grad_cache is None:
            if self.bypass_objective_jacobian:
                n_grad_caches = 1
            else:
                n_grad_caches = cnt

            self._grad_cache = np.zeros(
                (n_grad_caches, self._dij.total_num_of_bixels), dtype=np.float64
            )
        else:
            self._grad_cache.fill(0.0)

        for q in self._quantities:
            for scen_ix in q_scenarios[q.identifier]:
                if self.bypass_objective_jacobian:
                    cache_ix = 0
                else:
                    cache_ix = self._objectives_per_quantity[q.identifier]

                self._grad_cache[cache_ix, :] += (
                    q.compute_chain_derivative(self._grad_cache_intermediate[q.identifier], x)
                    .flat[scen_ix]
                    .squeeze()
                )

        return self._grad_cache

    def _objective_gradient(self, x: NDArray) -> NDArray:
        t = time.time()
        jac = np.sum(self._objective_jacobian(x), axis=0)
        self._deriv_times.append(time.time() - t)
        return jac

    def _objective_hessian(self, x: NDArray) -> NDArray:
        """Define the objective hessian."""
        return {}

    def _constraint_functions(self, x: NDArray) -> NDArray:
        """Define the constraint functions."""
        return None

    def _constraint_jacobian(self, x: NDArray) -> NDArray:
        """Define the constraint jacobian."""
        return None

    def _constraint_jacobian_structure(self) -> NDArray:
        """Define the constraint jacobian structure."""
        return None

    def _variable_bounds(self, x: NDArray) -> NDArray:
        """Define the variable bounds."""
        return {}

    def _solve(self) -> tuple[NDArray, dict]:
        """Solve the problem."""

        self._deriv_times = []
        self._obj_times = []

        x0 = np.zeros((self._dij.total_num_of_bixels,), dtype=np.float64)
        t = time.time()
        result = self.solver.solve(x0)
        self._solve_time = time.time() - t

        logger.info(
            "%d Objective function evaluations, avg. time: %g +/- %g s",
            len(self._obj_times),
            np.mean(self._obj_times),
            np.std(self._obj_times),
        )
        logger.info(
            "%d Derivative evaluations, avg. time: %g +/- %g s",
            len(self._deriv_times),
            np.mean(self._deriv_times),
            np.std(self._deriv_times),
        )
        logger.info("Solver time: %g s", self._solve_time)

        return result



from ._base import DoseEngineBase
from ._svdpb import PhotonPencilBeamSVDEngine
from ._hongpb import ParticleHongPencilBeamEngine

from ._factory import get_engine, get_available_engines, register_engine

register_engine(PhotonPencilBeamSVDEngine)
register_engine(ParticleHongPencilBeamEngine)

__all__ = [
    "DoseEngineBase",
    "PhotonPencilBeamSVDEngine",
    "ParticleHongPencilBeamEngine",
    "get_engine",
    "get_available_engines",
    "register_engine",
]



"""Module for handling CT images in pyRadPlan.

Defines a CT pydantic model and factories.
"""

import os
from abc import ABC
from typing import Annotated, Any, Union
from typing_extensions import Self
from pydantic import (
    Field,
    computed_field,
    model_validator,
)
import SimpleITK as sitk
import numpy as np
from pyRadPlan.core import resample_image

from pyRadPlan.core import PyRadPlanBaseModel, Grid


class CT(PyRadPlanBaseModel, ABC):
    """
    A class representing a CT (Computed Tomography) image.

    This class extends PyRadPlanBaseModel and provides functionality to
    handle CT images, including their properties like resolution, size, origin,
    and direction.

    Attributes
    ----------
    cube_hu : sitk.Image
        The CT image data in Hounsfield Units (HU).
    resolution : dict
        The resolution of the CT image in x, y, and z directions.
    size : tuple
        The size of the CT image in x, y, and z directions.
    origin : tuple
        The origin coordinates of the CT image.
    direction : tuple
        The direction cosines of the CT image.

    Methods
    -------
    to_matrad
        Converts the CT object to a dictionary compatible with matRad format.
    """

    cube_hu: Annotated[sitk.Image, Field(init=False, alias="cubeHU")]

    @model_validator(mode="before")
    @classmethod
    def validate_cube_hu(
        cls,
        data: Any,
    ) -> Any:
        """
        Validate and convert input data to SimpleITK image format.

        This method checks if the input data is in the correct format and
        converts numpy arrays to SimpleITK images if necessary. It also applies
        any specified image properties (origin, direction, spacing) to the
        SimpleITK image.

        Parameters
        ----------
        data : Any
            The input data to be validated and converted.

        Returns
        -------
        Any
            The validated and converted data.

        Raises
        ------
        ValueError
            If the HU cube is not present in the input dictionary.
        """

        if isinstance(data, CT):
            return data

        if isinstance(data, dict):  # check if data is dict
            # Check if the cube is contained in the data (either as field name or its alias)
            cube_hu = data.get("cube_hu", None)
            if cube_hu is None:
                cube_fieldname = cls.model_fields.get("cube_hu").validation_alias
                cube_hu = data.pop(cube_fieldname, None)

            if cube_hu is None:
                raise ValueError("HU cube not present in dictionary.")

            if isinstance(cube_hu, sitk.Image) and "origin" not in data:
                data["origin"] = cube_hu.GetOrigin()

            if isinstance(cube_hu, np.ndarray):
                # Now we do check if it is matRad data
                if data.get("cubeDim", None) is not None:
                    permute = True
                else:
                    permute = False

                # We have an array of multiple CT scenarios as numpy arrays
                if cube_hu.dtype == object:
                    num_cubes = len(cube_hu)

                    # If there are multiple cubes, we need to store them in a list
                    # for later use

                    if permute:
                        ct_scenarios = [
                            sitk.GetImageFromArray(np.transpose(cube_hu[i], (2, 0, 1)), False)
                            for i in range(num_cubes)
                        ]
                    else:
                        ct_scenarios = [
                            sitk.GetImageFromArray(cube_hu[i], False) for i in range(num_cubes)
                        ]

                    cube_hu = sitk.JoinSeries(ct_scenarios)
                else:
                    if permute:
                        cube_hu = sitk.GetImageFromArray(np.transpose(cube_hu, (2, 0, 1)), False)
                    else:
                        cube_hu = sitk.GetImageFromArray(cube_hu, False)
                    num_cubes = 1

            # Reassign the updated cube_hu to the data dictionary
            data["cube_hu"] = cube_hu

            # From here on we want an sitk.Image
            if not isinstance(data["cube_hu"], sitk.Image):
                raise ValueError(f"Unsupported format of HU cube: {type(data['cube_hu'])}")

            is4d = cube_hu.GetDimension() == 4
            # Now we parse the rest of the data if additional data is available in the dictionary

            if "direction" in data:
                data["cube_hu"].SetDirection(data["direction"])

            if "resolution" in data and all(key in data["resolution"] for key in ("x", "y", "z")):
                resolution = data["resolution"]
                if cube_hu.GetDimension() == 4:
                    cube_hu.SetSpacing([resolution["x"], resolution["y"], resolution["z"], 1.0])
                else:
                    cube_hu.SetSpacing([resolution["x"], resolution["y"], resolution["z"]])

            # TODO: Either set up a Grid or check for x/y/z to do the data management
            if "origin" in data:
                data["cube_hu"].SetOrigin(data["origin"])

            else:
                data["cube_hu"].SetOrigin(
                    -np.array(data["cube_hu"].GetSize())
                    / 2.0
                    * np.array(data["cube_hu"].GetSpacing())
                )
                if is4d:
                    data["cube_hu"].SetOrigin(np.append(data["cube_hu"].GetOrigin(), [0.0]))
            # elif all(key in data for key in ("x", "y", "z")):
            #     origin = np.array([data["x"][0], data["y"][0], data["z"][0]], dtype=float)
            #     if is4d:
            #         origin = np.append(origin, [0.0])
            #     data["cube_hu"].SetOrigin(origin)
        return data

    @computed_field
    @property
    def resolution(self) -> dict:
        """
        Get the resolution of the CT image.

        Returns
        -------
        dict
            A dictionary containing the resolution in x, y, and z directions.
        """

        resolution = self.cube_hu.GetSpacing()
        res = {}
        res["x"] = resolution[0]
        res["y"] = resolution[1]
        res["z"] = resolution[2]
        return res

    @computed_field
    @property
    def size(self) -> tuple:
        """
        Get the size of the CT image.

        Returns
        -------
        tuple
            A tuple containing the size in x, y, and z directions.
        """
        return self.cube_hu.GetSize()[0:3]

    @computed_field
    @property
    def cube_dim(self) -> tuple:
        """
        Get the size of the CT image.

        Returns
        -------
        tuple
            A tuple containing the size in x, y, and z directions.
        """
        return self.size[0:3]

    @computed_field
    @property
    def origin(self) -> tuple:
        """
        Get the origin of the CT image.

        Returns
        -------
        tuple
            A tuple containing the origin coordinates.
        """
        origin = self.cube_hu.GetOrigin()
        return origin[0:3]

    @computed_field
    @property
    def direction(self) -> tuple:
        """
        Get the direction of the CT image.

        Returns
        -------
        tuple
            A tuple containing the direction cosines.
        """
        direction = self.cube_hu.GetDirection()
        return direction

    @computed_field
    @property
    def num_of_ct_scen(self) -> int:
        """
        Get the number of CT scenarios.

        Returns
        -------
        int
            The number of CT scenarios.
        """
        if self.cube_hu.GetDimension() == 4:
            return self.cube_hu.GetSize()[3]
        return 1

    @computed_field
    @property
    def x(self) -> np.ndarray:
        """
        Calculate and get the x-vector of the CT image.

        Returns
        -------
        int
            The corresponding x-vector of the CT image.
        """

        x_spacing = self.cube_hu.GetSpacing()[0]
        origin = self.cube_hu.GetOrigin()[0]
        x = np.arange(origin, (self.size[0] * x_spacing + origin), x_spacing)
        return x

    @computed_field
    @property
    def y(self) -> np.ndarray:
        """
        Calculate and get the y-vector of the CT image.

        Returns
        -------
        int
            The corresponding y-vector of the CT image.
        """

        y_spacing = self.cube_hu.GetSpacing()[1]
        origin = self.cube_hu.GetOrigin()[1]
        y = np.arange(origin, (self.size[1] * y_spacing + origin), y_spacing)
        return y

    @computed_field
    @property
    def z(self) -> np.ndarray:
        """
        Calculate and get the z-vector of the CT image.

        Returns
        -------
        int
            The corresponding z-vector of the CT image.
        """

        z_spacing = self.cube_hu.GetSpacing()[2]
        origin = self.cube_hu.GetOrigin()[2]
        z = np.arange(origin, (self.size[2] * z_spacing + origin), z_spacing)
        return z

    @property
    def grid(self) -> Grid:
        """
        Get the grid of the CT image.

        Returns
        -------
        Grid
            The grid of the CT image.
        """
        return Grid.from_sitk_image(self.cube_hu)

    def world_to_cube_coords(self, world_coords: np.ndarray) -> np.ndarray:
        """
        Convert world coordinates to cube coordinates.

        Parameters
        ----------
        world_coords : np.ndarray
            The world coordinates to convert.

        Returns
        -------
        np.ndarray
            The converted cube coordinates.
        """

        return np.array(self.cube_hu.TransformPhysicalPointToIndex(world_coords))

    def compute_wet(self, hlut: np.ndarray) -> sitk.Image:
        """
        Compute the water equivalent thickness (WET).

        Uses a provided appropriate Hounsfield Look-Up Table (HLUT).

        Parameters
        ----------
        hlut : np.ndarray
            The Hounsfield Look-Up Table (HLUT) to compute the WET.

        Returns
        -------
        sitk.Image
            The water equivalent thickness (WET) image.
        """

        if hlut.shape[1] != 2 and hlut.shape[0] <= 1:
            raise ValueError(
                "HLUT must have 2 columns of values, with the first column being the HU values "
                "and the second column being the WET values with each at least 2 values."
            )

        # Sort the HLUT by HU values
        ix = np.argsort(hlut[:, 0])
        xp = hlut[ix, 0]
        fp = hlut[ix, 1]

        # Get the np view of the image
        ct_array = sitk.GetArrayViewFromImage(self.cube_hu)

        # Convert the WET array back to a SimpleITK image
        wet_image = sitk.GetImageFromArray(np.interp(ct_array, xp, fp))
        wet_image.SetOrigin(self.cube_hu.GetOrigin())
        wet_image.SetSpacing(self.cube_hu.GetSpacing())
        wet_image.SetDirection(self.cube_hu.GetDirection())

        return wet_image

    def to_matrad(self, context: str = "mat-file") -> Any:
        """
        Convert the CT object to a dictionary compatible with matRad format.

        Returns
        -------
        Any
            A dictionary containing the CT data in matRad format.
        """
        ct_dict = super().to_matrad(context=context)
        # as cubeHU for the matRad format
        sitk_image = ct_dict["cubeHU"]
        # Flip axes for matRad format
        numpy_array = np.transpose(sitk.GetArrayFromImage(sitk_image), (1, 2, 0))
        # We also need to put the image array into a ndarray of objects
        ct_dict["cubeHU"] = np.ndarray(shape=(1,), dtype=object)
        ct_dict["cubeHU"][0] = numpy_array
        ct_dict["cubeDim"] = np.array(ct_dict["cubeDim"], dtype=float)
        return ct_dict

    def resample_to_grid(self, grid: Grid) -> Self:
        """
        Resample the CT image to match the specified grid.

        Parameters
        ----------
        grid : Grid
            The grid to resample the CT image to.

        Returns
        -------
        Self
            The resampled CT object.
        """

        resampler = sitk.ResampleImageFilter()
        resampler.SetInterpolator(sitk.sitkLinear)
        resampler.SetOutputDirection(grid.direction.ravel())
        resampler.SetOutputOrigin(grid.origin)
        resampler.SetOutputSpacing(grid.resolution_vector)
        resampler.SetSize(grid.dimensions)

        if self.num_of_ct_scen > 1:
            new_ct = []
            for i in range(self.cube_hu.GetSize()[-1]):
                new_ct.append(resampler.Execute(self.cube_hu[:, :, :, i]))
            new_ct = sitk.JoinSeries(new_ct)
        else:
            new_ct = resampler.Execute(self.cube_hu)

        return self.model_validate({"cube_hu": new_ct})


def resample_ct(ct: CT, interpolator=sitk.sitkBSpline, **kwargs) -> CT:
    """
    Resample CT to a different grid.

    Returns
    -------
    CT
        A resampled CT object
    """

    ct_cube = ct.cube_hu

    resampled_ct_cube = resample_image(input_image=ct_cube, interpolator=interpolator, **kwargs)

    return validate_ct(cube_hu=resampled_ct_cube)


def ct_from_file(file_path: Union[str, os.PathLike]) -> CT:
    """
    Create a CT object from a file.

    Parameters
    ----------
    file_path : Union[str, os.PathLike]
        The path to the CT image file.

    Returns
    -------
    CT
        A CT object created from the file.

    Raises
    ------
    ValueError
        If reading DICOM series from folders is attempted (not implemented).
    """
    if os.path.isfile(file_path):
        sitk_image = sitk.ReadImage(file_path)
        # GetArrayFromImage accounts for the difference between numpy (z, x, y)
        # and SITK (x, y, z) standards

    else:
        raise ValueError("Reading dicom series from folders not implemented")

    return create_ct(cube_hu=sitk_image)


def create_ct(data: Union[dict[str, Any], CT, os.PathLike, str, None] = None, **kwargs) -> CT:
    """
    Create a CT object from various input types.

    Parameters
    ----------
    data : Union[dict[str, Any], CT, os.PathLike, str, None], optional
        The input data to create the CT object from. Can be a dictionary,
        existing CT object, file path, or None.
    **kwargs
        Additional keyword arguments to create the CT object.

    Returns
    -------
    CT
        A CT object created from the input data or keyword arguments.
    """
    if data:  # If data is not None
        # If data is already a CT object return it directly
        if isinstance(data, CT):
            return data
        # If data is in a file get CT object from file
        if isinstance(data, str) or isinstance(data, os.PathLike):
            return ct_from_file(data)
        # If data is in a dictionary create an SimpleITK image and then
        # the CT object
        return CT.model_validate(data)
    # If neither CT object nor dictionary try to
    # get model from keyword arguments
    return CT(**kwargs)


def validate_ct(ct: Union[dict[str, Any], CT, os.PathLike, None] = None, **kwargs) -> CT:
    """
    Validate and create a CT object.

    This function is a wrapper around create_ct and ensures the returned object
    is a valid CT instance.

    Parameters
    ----------
    ct : Union[dict[str, Any], CT, os.PathLike, None], optional
        The input data to validate and create the CT object from.
    **kwargs
        Additional keyword arguments to create the CT object.

    Returns
    -------
    CT
        A validated CT object.
    """
    return create_ct(ct, **kwargs)



from numpy.typing import NDArray
import pint

from pyRadPlan.quantities._base import FluenceDependentQuantity

ureg = pint.UnitRegistry()


class LETxDose(FluenceDependentQuantity):
    """LETxDose quantity depending on fluence."""

    unit = ureg.gray * ureg.micrometer / ureg.keV
    dim = 1
    identifier = "let_dose"
    name = "LETxDose"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.identifier not in self._dij.model_fields or self._dij.let_dose is None:
            raise ValueError(f"Quantity {self.identifier} not available in Dij object.")

    def _compute_quantity_single_scenario_from_cache(self, scenario_index: int) -> NDArray:
        return self._dij.let_dose.flat[scenario_index] @ self._w_cache

    def _compute_chained_fluence_gradient_single_scenario_from_cache(
        self, d_quantity, scenario_index: int
    ) -> NDArray:
        return d_quantity @ self._dij.let_dose.flat[scenario_index]



import pytest

from pyRadPlan.optimization.solvers import get_available_solvers, get_solver, SolverBase


def test_get_available_solvers():
    solvers = get_available_solvers()
    assert isinstance(solvers, dict)
    assert len(solvers) > 0
    assert all([isinstance(k, str) for k in solvers.keys()])
    assert all([issubclass(v, SolverBase) for v in solvers.values()])


def test_get_solver():
    solvers = get_available_solvers()
    for solver_name, solver_class in solvers.items():
        solver = get_solver(solver_name)
        assert isinstance(solver, solver_class)
        assert isinstance(solver, SolverBase)
        assert solver.short_name == solver_name
        assert solver.name == solver_class.name
        assert solver.short_name == solver_class.short_name


def test_get_solver_invalid():
    with pytest.raises(KeyError):
        get_solver("invalid_solver")
    with pytest.raises(ValueError):
        get_solver(123)
    with pytest.raises(NotImplementedError):
        get_solver({"solver": "scipy"})  # valid solver, but dict not implemented yet



import pytest
import SimpleITK as sitk
import numpy as np
import scipy.sparse as sp
from pyRadPlan.core import Grid
from pyRadPlan.dij import Dij, create_dij, validate_dij, compose_beam_dijs


@pytest.fixture
def sample_dij_dict():
    dose_information = {
        "ct_grid": {
            "resolution": {"x": 1.5, "y": 1.5, "z": 1.5},
            "dimensions": (334, 334, 214),
            "num_of_voxels": 334 * 334 * 214,
        },
        "dose_grid": {
            "resolution": {"x": 3.0, "y": 3.0, "z": 3.0},
            "dimensions": (167, 167, 107),
            "num_of_voxels": 167 * 167 * 107,
        },
        "physical_dose": [[[sp.csc_matrix((167 * 167 * 107, 6506), dtype=float)]]],
        "num_of_beams": 1,
        "total_num_of_bixels": 6506,
        "bixel_num": np.ones(6506),
        "ray_num": np.ones(6506),
        "beam_num": np.ones(6506),
    }
    return dose_information


@pytest.fixture
def sample_dij_dict_camel():
    dose_information = {
        "ctGrid": {
            "resolution": {"x": 3.0, "y": 3.0, "z": 3.0},
            "dimensions": (167, 167, 107),
            "numOfVoxels": 167 * 167 * 107,
        },
        "doseGrid": {
            "resolution": {"x": 3.0, "y": 3.0, "z": 3.0},
            "dimensions": (167, 167, 107),
            "numOfVoxels": 167 * 167 * 107,
        },
        "physicalDose": [[[sp.csc_matrix((167 * 167 * 107, 6506), dtype=float)]]],
        "numOfBeams": 1,
        "totalNumOfBixels": 6506,
        "bixelNum": np.ones(6506),
        "rayNum": np.ones(6506),
        "beamNum": np.ones(6506),
    }
    return dose_information


def test_dij_creation_from_dict(sample_dij_dict):
    dij = create_dij(**sample_dij_dict)
    assert isinstance(dij, Dij)
    assert isinstance(dij.dose_grid, Grid)
    assert isinstance(dij.ct_grid, Grid)
    assert isinstance(dij.physical_dose, np.ndarray)
    assert isinstance(dij.physical_dose.flat[0], sp.csc_matrix)
    assert dij.total_num_of_bixels == 6506
    assert dij.num_of_voxels == 167 * 167 * 107
    assert dij.bixel_num.shape == (6506,)
    assert dij.ray_num.shape == (6506,)
    assert dij.beam_num.shape == (6506,)


def test_dij_creation_from_camel_dict(sample_dij_dict_camel):
    dij = validate_dij(**sample_dij_dict_camel)
    assert isinstance(dij, Dij)
    assert isinstance(dij.dose_grid, Grid)
    assert isinstance(dij.ct_grid, Grid)
    assert isinstance(dij.physical_dose, np.ndarray)
    assert isinstance(dij.physical_dose.flat[0], sp.csc_matrix)
    assert dij.total_num_of_bixels == 6506
    assert dij.num_of_voxels == 167 * 167 * 107


def test_dij_creation_invalid_physical_dose_shape(sample_dij_dict):
    sample_dij_dict["physical_dose"] = sp.csr_matrix((100, 4), dtype=float)  # Invalid shape
    with pytest.raises(ValueError):
        create_dij(**sample_dij_dict)


def test_dij_creation_invalid_beam_num(sample_dij_dict):
    sample_dij_dict["beam_num"] = np.ones(5)  # Invalid shape
    with pytest.raises(ValueError):
        create_dij(**sample_dij_dict)


def test_dij_creation_invalid_ray_num(sample_dij_dict):
    sample_dij_dict["ray_num"] = np.ones(5)  # Invalid shape
    with pytest.raises(ValueError):
        create_dij(**sample_dij_dict)


def test_dij_creation_invalid_bixel_num(sample_dij_dict):
    sample_dij_dict["bixel_num"] = np.ones(5)  # Invalid shape
    with pytest.raises(ValueError):
        create_dij(**sample_dij_dict)


def test_dij_creation_invalid_unique_indices_in_beam_num(sample_dij_dict):
    # number_of_beams = 1
    sample_dij_dict["beam_num"][3000:] = 2
    with pytest.raises(ValueError):
        create_dij(**sample_dij_dict)


def test_dij_creation_invalid_number_of_beams(sample_dij_dict):
    # beam_num only ones
    sample_dij_dict["num_of_beams"] = 2
    with pytest.raises(ValueError):
        create_dij(**sample_dij_dict)


def test_dij_to_matrad(sample_dij_dict):
    dij = create_dij(**sample_dij_dict)
    matrad_dict = dij.to_matrad()
    assert isinstance(matrad_dict, dict)
    assert "ctGrid" in matrad_dict
    assert "doseGrid" in matrad_dict
    assert "totalNumOfBixels" in matrad_dict
    assert isinstance(matrad_dict["totalNumOfBixels"], int)
    assert isinstance(matrad_dict["ctGrid"], dict)
    assert isinstance(matrad_dict["doseGrid"], dict)
    assert matrad_dict["totalNumOfBixels"] == 6506


def test_dij_to_matrad_from_csc_matrix(sample_dij_dict):
    sample_dij_dict2 = sample_dij_dict
    sample_dij_dict2["physical_dose"][0][0][0] = sp.csc_matrix(
        (167 * 167 * 107, 6506), dtype=float
    )
    dij = create_dij(**sample_dij_dict2)
    matrad_dict = dij.to_matrad()
    assert isinstance(matrad_dict, dict)
    assert isinstance(matrad_dict["physicalDose"].flat[0], sp.csc_matrix)


def test_dij_to_matrad_from_csc_array(sample_dij_dict):
    sample_dij_dict["physical_dose"][0][0][0] = sp.csc_array((167 * 167 * 107, 6506), dtype=float)
    dij = create_dij(**sample_dij_dict)
    matrad_dict = dij.to_matrad()
    assert isinstance(matrad_dict, dict)
    assert isinstance(matrad_dict["physicalDose"].flat[0], sp.csc_matrix)


def test_dij_to_matrad_from_csr_array(sample_dij_dict):
    sample_dij_dict["physical_dose"][0][0][0] = sp.csr_array((167 * 167 * 107, 6506), dtype=float)
    dij = create_dij(**sample_dij_dict)
    matrad_dict = dij.to_matrad()
    assert isinstance(matrad_dict, dict)
    assert isinstance(matrad_dict["physicalDose"].flat[0], sp.csc_matrix)


def test_create_dij_from_grids():
    ct_dict = {
        "resolution": {"x": 1, "y": 1, "z": 1},
        "dimensions": (167, 167, 107),
    }
    dose_dict = {
        "resolution": {"x": 1, "y": 1, "z": 1},
        "dimensions": (167, 167, 107),
    }
    ct_grid = Grid.model_validate(ct_dict)
    dose_grid = Grid.model_validate(dose_dict)

    sample_dij_dict = {
        "physical_dose": sp.csr_matrix((167 * 167 * 107, 6506), dtype=float),
        "num_of_beams": 1,
        "total_num_of_bixels": 6506,
        "bixelNum": np.ones(6506),
        "rayNum": np.ones(6506),
        "beamNum": np.ones(6506),
    }
    create_dij(ct_grid=ct_grid, dose_grid=dose_grid, **sample_dij_dict)


def test_result_computation_physical_dose_dose_grid(sample_dij_dict):
    dij = create_dij(**sample_dij_dict)
    f = np.ones((dij.total_num_of_bixels,), dtype=np.float32)
    result = dij.compute_result_dose_grid(f)
    assert isinstance(result, dict)
    assert "physical_dose" in result
    assert isinstance(result["physical_dose"], sitk.Image)
    assert result["physical_dose"].GetSize() == dij.dose_grid.dimensions


def test_result_computation_physical_dose_ct_grid(sample_dij_dict):
    dij = create_dij(**sample_dij_dict)
    f = np.ones((dij.total_num_of_bixels,), dtype=np.float32)
    result = dij.compute_result_ct_grid(f)
    assert isinstance(result, dict)
    assert "physical_dose" in result
    assert isinstance(result["physical_dose"], sitk.Image)
    assert result["physical_dose"].GetSize() == dij.ct_grid.dimensions


def test_result_computation_let(sample_dij_dict):
    dij = create_dij(**sample_dij_dict)
    f = np.zeros((dij.total_num_of_bixels,), dtype=np.float32)

    dij.let_dose = dij.physical_dose

    result = dij.compute_result_dose_grid(f)

    assert isinstance(result, dict)
    assert "let" in result
    assert "physical_dose" in result
    assert isinstance(result["let"], sitk.Image)
    assert isinstance(result["physical_dose"], sitk.Image)
    assert result["let"].GetSize() == dij.dose_grid.dimensions
    assert result["physical_dose"].GetSize() == dij.dose_grid.dimensions

    with pytest.raises(ValueError):
        dij.physical_dose = None
        dij.compute_result_dose_grid(f)


def test_result_computation_biodose(sample_dij_dict):
    dij = create_dij(**sample_dij_dict)
    f = np.zeros((dij.total_num_of_bixels,), dtype=np.float32)

    dij.alpha_dose = dij.physical_dose
    dij.sqrt_beta_dose = dij.physical_dose

    result = dij.compute_result_dose_grid(f)

    assert isinstance(result, dict)
    assert "effect" in result
    assert "physical_dose" in result
    assert isinstance(result["effect"], sitk.Image)
    assert isinstance(result["physical_dose"], sitk.Image)
    assert result["effect"].GetSize() == dij.dose_grid.dimensions
    assert result["physical_dose"].GetSize() == dij.dose_grid.dimensions

    with pytest.raises(ValueError):
        dij.physical_dose = None
        dij.compute_result_dose_grid(f)


def test_compose_beam_dijs(sample_dij_dict):
    dij = create_dij(**sample_dij_dict)
    composed_dij = compose_beam_dijs([dij, dij])
    assert isinstance(composed_dij, Dij)
    assert composed_dij.num_of_beams == 2 * dij.num_of_beams
    assert composed_dij.total_num_of_bixels == 2 * dij.total_num_of_bixels
    assert composed_dij.beam_num.shape == (2 * dij.total_num_of_bixels,)
    assert composed_dij.ray_num.shape == (2 * dij.total_num_of_bixels,)
    assert composed_dij.bixel_num.shape == (2 * dij.total_num_of_bixels,)
    assert composed_dij.dose_grid == dij.dose_grid
    assert composed_dij.ct_grid == dij.ct_grid



"""Structure Set Implementation."""

from typing import Any, Union
from typing_extensions import Self
from pydantic import Field, model_validator

import numpy as np
from scipy import ndimage
import SimpleITK as sitk

from pyRadPlan.core import PyRadPlanBaseModel
from pyRadPlan.cst import VOI, ExternalVOI, validate_voi
from pyRadPlan.ct import CT, validate_ct


class StructureSet(PyRadPlanBaseModel):
    """Represents a Structure Set for a Patient."""

    vois: list[VOI] = Field(init=False, description="List of VOIs in the Structure Set")
    ct_image: CT = Field(init=False, description="Reference to the CT Image")

    @model_validator(mode="after")
    def check_cst(self) -> Self:
        """Check if the VOIs are valid and reference the same CT."""

        if isinstance(self.vois, list):
            for voi in self.vois:
                if voi.ct_image != self.ct_image:
                    raise ValueError("All VOIs must reference the same CT image.")

        return self

    def to_matrad(self, context: str = "mat-file") -> Any:
        """Convert the StructureSet to a matRad writeable format."""

        if context != "mat-file":
            raise ValueError(f"Context {context} not supported")

        export_cell_list = []
        for i, voi in enumerate(self.vois):
            voi_list = voi.to_matrad(context=context)
            voi_list[0] = i
            # TODO: set objectives here
            voi_list[5] = {}

            export_cell_list.append(voi_list)

        return export_cell_list

    # Additional Properties
    @property
    def voi_types(self) -> list:
        """Return the unique VOI types in the Structure Set."""
        return list({voi.voi_type for voi in self.vois})

    def target_union_voxels(self, order="sitk") -> np.ndarray:
        """Return the union of all target indices."""
        target_indices = []
        for voi in self.vois:
            if voi.voi_type == "TARGET":
                target_indices.append(voi.get_indices(order=order))

        return np.unique(np.concatenate(target_indices))

    def target_union_mask(self) -> sitk.Image:
        """Return the union mask of all target indices."""

        target_indices = self.target_union_voxels(order="numpy")

        # Creates a copy of the CT image with all zeros

        if self.ct_image.cube_hu.GetDimension() == 4:
            sz = np.array(self.ct_image.cube_hu.GetSize())
            sz[3] = 0
            tmpmask3d = sitk.Extract(
                self.ct_image.cube_hu,
                size=sz.tolist(),
                index=[0, 0, 0, 0],
                directionCollapseToStrategy=sitk.ExtractImageFilter.DIRECTIONCOLLAPSETOSUBMATRIX,
            )

        else:
            tmpmask3d = self.ct_image.cube_hu

        mask = sitk.GetArrayViewFromImage(tmpmask3d).astype(np.uint8)
        mask.fill(0)
        mask.ravel()[target_indices] = 1

        mask_image = sitk.GetImageFromArray(mask)
        mask_image.CopyInformation(tmpmask3d)

        return mask_image

    def patient_voxels(self, order="sitk") -> np.ndarray:
        """Return the union of all patient indices."""

        # First check if we have an "EXTERNAL" VOI designating the outer contour
        for voi in self.vois:
            if isinstance(voi, ExternalVOI):
                return voi.get_indices(order=order)

        patient_indices = []
        for voi in self.vois:
            patient_indices.append(voi.get_indices(order=order))

        return np.unique(np.concatenate(patient_indices))

    def patient_mask(self) -> sitk.Image:
        """Return the union mask of all patient contours (or the EXTERNAL
        contour if provided).
        """

        patient_indices = self.patient_voxels(order="numpy")

        # Creates a copy of the CT image with all zeros
        mask = sitk.GetArrayFromImage(self.ct_image.cube_hu).astype(np.uint8)
        mask.fill(0)
        mask.ravel()[patient_indices] = 1

        mask_image = sitk.GetImageFromArray(mask)
        mask_image.CopyInformation(self.ct_image.cube_hu)

        return mask_image

    def target_center_of_mass(self) -> np.ndarray:
        """Return the center of mass of the target."""
        mask_image = self.target_union_mask()
        mask = sitk.GetArrayViewFromImage(
            mask_image
        ).T  # Transpose allows use to use sitk indexing

        if mask.ndim == 4:
            mask = mask[:, :, :, 0]

        cm_index = ndimage.center_of_mass(mask)

        cm = mask_image.TransformContinuousIndexToPhysicalPoint(cm_index)

        return np.array(cm)

    def resample_on_new_ct(self, new_ct: CT) -> Self:
        """
        Resample the StructureSet on a new CT.

        Parameters
        ----------
        new_ct : CT
            The new CT to resample the StructureSet on.

        Returns
        -------
        StructureSet
            The resampled StructureSet.
        """

        new_model = self.model_dump()

        if new_model["ct_image"] != new_ct:
            new_model["ct_image"] = new_ct
            new_model["vois"] = [voi.resample_on_new_ct(new_ct) for voi in self.vois]
        return self.model_validate(new_model)

    def apply_overlap_priorities(self) -> Self:
        """
        Apply overlaps to the StructureSet.

        Returns
        -------
        StructureSet
            The StructureSet with overlaps applied.
        """

        # gather overlaps
        overlaps = [v.overlap_priority for v in self.vois]

        # sort by overlap priority
        ix_sorted = np.argsort(overlaps)

        overlap_mask = self.vois[ix_sorted[0]].mask
        new_vois = [None] * len(self.vois)
        new_vois[ix_sorted[0]] = self.vois[ix_sorted[0]].copy()

        for i, ix_voi in enumerate(ix_sorted[1:], 1):
            curr_voi = self.vois[ix_voi].copy()
            curr_mask = curr_voi.mask

            # if the overlap prirority is higher than we need to apply overlap
            if curr_voi.overlap_priority > new_vois[ix_sorted[i - 1]].overlap_priority:
                curr_mask = sitk.MaskNegated(curr_mask, overlap_mask)

            curr_voi.mask = curr_mask

            overlap_mask = sitk.Or(overlap_mask, curr_mask)

            # sitk.Show(overlap_mask, debugOn = True)

            new_vois[ix_voi] = curr_voi

        return self.model_copy(deep=True, update={"vois": new_vois})


def create_cst(
    cst_data: Union[dict[str, Any], StructureSet, None] = None,
    ct: Union[CT, dict, None] = None,
    **kwargs,
) -> StructureSet:
    """
    Create a StructureSet from various input types.

    Parameters
    ----------
    cst_data : Union[dict[str, Any], StructureSet, None] , optional
        The input data to create the CT object from. Can be a dictionary,
        existing CT object, file path, or None.

    ct : Union[CT,dict,None], CT, None], optional
        The input data to create the CT object from. Can be a dictionary,
        existing CT object, or None.

    **kwargs
        Additional keyword arguments to create the CT object.

    Returns
    -------
    cst
        A StructureSet object created from the input data or keyword arguments.
    """

    # Check if already a valid model
    if isinstance(cst_data, StructureSet):
        if ct is not None and cst_data.ct_image != ct:
            raise ValueError("CT image mismatch between StructureSet and provided CT")
        return cst_data

    # validate ct if present
    if ct is not None:
        ct = validate_ct(ct)

    # If already a model dictionary check the ct setup
    if isinstance(cst_data, dict):
        ct_image = cst_data.pop("ct_image", cst_data.pop("ctImage", None))
        if ct_image is not None:
            cst_data["ct_image"] = ct_image
            if ct is not None and ct != ct_image:
                raise ValueError("CT image mismatch between StructureSet and provided CT")
        elif ct is not None:
            cst_data["ct_image"] = ct
        else:
            raise ValueError("No CT image reference provided!")
        return StructureSet.model_validate(cst_data)

    if cst_data is None and ct is not None:  # If data is None
        return StructureSet(ct_image=ct, **kwargs)
    if cst_data is None and ct is None:
        return StructureSet(**kwargs)

    # Other methods need the CT
    if ct is None:
        raise ValueError("No CT image reference provided!")

    # Creation from an nd array (cell array matRad format)
    if isinstance(cst_data, np.ndarray) and cst_data.dtype == object:
        cst_data = cst_data.to_list()

    # a list of volume information (e.g. imported from pymatreader from matRad mat file)
    if isinstance(cst_data, list):
        voi_list = []
        for vdata in cst_data:
            # First try to read the index lists
            idx_list = []
            # Only one scenario (3D CT)
            if not isinstance(vdata[3], list):
                idx_list.append(vdata[3].astype(int).tolist())
            # Multiple scenarios (4D CT)
            else:
                for vdata_scen in vdata[3]:
                    idx_list.append(vdata_scen.astype(int).tolist())

            # Now we create isimple ITK masks
            masks = []
            for idx in idx_list:
                # TODO: Check index ordering
                tmp_mask = np.zeros((ct.size[2], ct.size[0], ct.size[1]), dtype=np.uint8)
                tmp_mask.flat[np.asarray(idx) - 1] = 1
                tmp_mask = np.swapaxes(tmp_mask, 1, 2)
                mask_image = sitk.GetImageFromArray(tmp_mask)
                mask_image.CopyInformation(ct.cube_hu)

                masks.append(mask_image)

            # For 4D, we need to join the masks. We also check here if the number of masks we could
            # extract matches the number of dimensions in the CT image
            if ct.cube_hu.GetDimension() == 4:
                # First check if the mask is the same for all 4D scenarios
                if len(masks) == 1:
                    masks = [masks[0] for _ in range(ct.cube_hu.GetDimension())]

                # Now do a sanity check that we don't have an incompatible number of masks
                if len(masks) != ct.cube_hu.GetSize()[3]:
                    raise ValueError("Incompatible number of masks for 4D CT")

                masks = sitk.JoinSeries(*masks)

            # If it is a 3D CT, we just drop the list
            elif ct.cube_hu.GetDimension() == 3:
                masks = masks[0]
            else:
                raise ValueError("Sanity Check failed -- unsupported CT dimensionality")

            # Check Objectives
            if len(vdata) > 5:
                objectives = vdata[5]
                if not isinstance(objectives, list):
                    objectives = [objectives]
            else:
                objectives = []

            voi = validate_voi(
                name=str(vdata[1]),
                voi_type=str(vdata[2]),
                mask=masks,
                ct_image=ct,
                objectives=objectives,
            )

            voi_list.append(voi)

        cst_dict = {"vois": voi_list, "ct_image": ct}
        return StructureSet.model_validate(cst_dict)

    raise ValueError("Invalid input data for creating a StructureSet.")


def validate_cst(
    cst_data: Union[dict[str, Any], StructureSet, None] = None,
    ct: Union[CT, dict, None] = None,
    **kwargs,
) -> StructureSet:
    """
    Validate StructureSet.

    Parameters
    ----------
    cst_data : Union[dict[str, Any], StructureSet, None] , optional
        The input data to create the CT object from. Can be a dictionary,
        existing CT object, file path, or None.

    ct : Union[CT,dict,None], CT, None], optional
        The input data to create the CT object from. Can be a dictionary,
        existing CT object, or None.

    **kwargs
        Additional keyword arguments to create the CT object.

    Returns
    -------
    cst
        A StructureSet object created from the input data or keyword arguments.
    """
    return create_cst(cst_data, ct, **kwargs)



"""Base class for pencil beam dose calculation algorithms."""

from abc import abstractmethod
from typing import Any, Literal
import warnings
import logging
import time
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

import SimpleITK as sitk
import numpy as np
from scipy import sparse

from pyRadPlan.core import resample_image, np2sitk
from pyRadPlan.ct import CT, default_hlut
from pyRadPlan.cst import StructureSet
from pyRadPlan.stf import SteeringInformation
from pyRadPlan.geometry import get_beam_rotation_matrix
from pyRadPlan.raytracer import RayTracerSiddon

from ._base import DoseEngineBase


logger = logging.getLogger(__name__)


class PencilBeamEngineAbstract(DoseEngineBase):
    """
    An abstract class representing the Pencil Beam Engine.

    This class extends DoseEngineBase and provides the foundational structure for implementing a
    pencil beam dose calculation engine. It includes methods for initializing dose calculations,
    setting defaults, and various helper methods required for the dose calculation process.

    Attributes
    ----------
    keep_rad_depth_cubes : bool
        Flag to keep radiation depth cubes.
    geometric_lateral_cutoff : float
        Lateral geometric cut-off in mm, used for raytracing and geometry.
    dosimetric_lateral_cutoff : float
        Relative dosimetric cut-off (in fraction of values calculated).
    ssd_density_threshold : float
        Threshold for SSD computation.
    use_given_eq_density_cube : bool
        Use the given density cube ct.cube and omit conversion from cubeHU.
    ignore_outside_densities : bool
        Ignore densities outside of cst contours.
    num_of_dij_fill_steps : int
        Number of times during dose calculation the temporary containers are moved to a sparse
        matrix.
    cube_wed : sitk.Image
        Relative electron density / stopping power cube.
    hlut : np.ndarray
        Hounsfield lookup table to create relative electron density cube.

    Methods
    -------
    __init__()
        Initializes the PencilBeamEngineAbstract class.
    set_defaults()
        Sets default values for the attributes.
    _compute_bixel(curr_ray, k)
        Abstract method to compute bixel.
    _calc_dose(ct, cst, stf)
        Calculates the dose.
    _init_dose_calc(ct, cst, stf)
        Initializes dose calculation.
    _allocate_quantity_matrices(dij, names)
        Allocates quantity matrix containers.
    _init_beam(curr_beam, ct, cst, stf, i)
        Initializes the beam.
    _init_ray(curr_beam, j)
        Initializes the ray.
    _extract_single_scenario_ray(ray, scen_idx)
        Extracts a single scenario ray.
    _get_ray_geometry_from_beam(ray, curr_beam)
        Gets ray geometry from beam.
    _get_lateral_distance_from_dose_cutoff_on_ray(ray)
        Gets lateral distance from dose cutoff on ray.
    _fill_dij(bixel, dij, stf, scen_idx, curr_beam_idx, curr_ray_idx, curr_bixel_idx, counter)
        Fills the dose influence matrix (dij).
    _finalize_dose(dij)
        Finalizes the dose.
    calcGeoDists(rot_coords_bev, sourcePoint_bev, targetPoint_bev, SAD, radDepthIx, lateralCutOff)
        Calculates geometric distances.
    geometric_cutoff()
        Property for geometric cutoff.
    warn_deprecated_engine_property(old_name, new_name, internal_name)
        Warns about deprecated engine properties.
    """

    keep_rad_depth_cubes: bool
    geometric_lateral_cutoff: float
    dosimetric_lateral_cutoff: float
    ssd_density_threshold: float
    use_given_eq_density_cube: bool
    ignore_outside_densities: bool
    trace_on_dose_grid: bool
    cube_wed: sitk.Image
    hlut: np.ndarray

    def __init__(self, pln=None):
        self.keep_rad_depth_cubes = False
        self.geometric_lateral_cutoff: float = 50
        self.dosimetric_lateral_cutoff: float = 0.9950
        self.ssd_density_threshold: float = 0.0500
        self.use_given_eq_density_cube: bool = False
        self.ignore_outside_densities: bool = True
        self.trace_on_dose_grid: bool = True
        self.cube_wed = None
        self.hlut = None

        self._computed_quantities = []
        self._effective_lateral_cutoff = None
        self._num_of_bixels_container = None
        self._rad_depth_cubes = []
        self._raytracer = None

        super().__init__(pln)

    @abstractmethod
    def _compute_bixel(self, curr_ray, k):
        raise NotImplementedError("Method _compute_bixel must be implemented in derived class.")

    def _calc_dose(self, ct: CT, cst: StructureSet, stf: SteeringInformation):
        """
        Calculate the dose using the pencil beam method.

        Parameters
        ----------
        ct : CT
            The CT object.
        cst : StructureSet
            The structure set object.
        stf : SteeringInformation
            The steering information object.

        Returns
        -------
        dict
            The dose influence matrix dictionary.
        """

        # Initialize
        dij = self._init_dose_calc(ct, cst, stf)

        # We loop over scenario in the scenario model
        # TODO: we need to correctly work out scenarios
        for shift_scen in range(self.mult_scen.tot_num_shift_scen):
            # Find first instance of the shift to select the shift values
            # TODO!: Check for more than one Scenario
            ix_shift_scen = np.where(self.mult_scen.linear_mask[:, 1] == shift_scen)

            scen_stf = stf
            # Manipulate isocenter
            for beam in scen_stf.beams:
                beam.iso_center += self.mult_scen.iso_shift[ix_shift_scen, :].reshape(-1)

            if self.mult_scen.tot_num_shift_scen > 1:
                logger.info(
                    f"Shift scenario {shift_scen} of {self.mult_scen.tot_num_shift_scen}: \n"
                )

            bixel_counter = 0

            # Loop over all beams
            with logging_redirect_tqdm():
                for i in tqdm(range(dij["num_of_beams"]), desc="Beam", unit="b", leave=False):
                    # Initialize Beam Geometry
                    t = time.time()
                    curr_beam = self._init_beam(dij, ct, cst, scen_stf, i)
                    logger.info("Beam %d initialized in %f seconds.", i + 1, time.time() - t)

                    # Keep tabs on bixels computed in this beam
                    bixel_beam_counter = 0

                    # Ray calculation
                    for j in tqdm(
                        range(curr_beam["beam"]["num_of_rays"]), desc="Ray", unit="r", leave=False
                    ):
                        # Initialize Ray Geometry
                        curr_ray = self._init_ray(curr_beam, j)

                        # check if ray hit anything. If so, skip the computation
                        if all(not arr.size for arr in curr_ray["rad_depths"]):
                            continue

                        # TODO: incorporate scenarios correctly
                        for ct_scen in range(self.mult_scen.num_of_ct_scen):
                            for range_scen in range(self.mult_scen.tot_num_range_scen):
                                # Obtain scenario index
                                full_scen_idx = self.mult_scen.sub2scen_ix(
                                    ct_scen, shift_scen, range_scen
                                )

                                if self.mult_scen.scen_mask[full_scen_idx]:
                                    # Extract single scenario ray
                                    scen_ray = self._extract_single_scenario_ray(
                                        curr_ray, full_scen_idx
                                    )

                                    for k in range(curr_ray["num_of_bixels"]):
                                        # Bixel Computation
                                        curr_bixel = self._compute_bixel(scen_ray, k)

                                        # fill the current bixel in the sparse dose influence
                                        # matrix
                                        self._fill_dij(
                                            curr_bixel,
                                            dij,
                                            scen_stf,
                                            full_scen_idx,
                                            i,
                                            j,
                                            k,
                                            bixel_counter + k,
                                        )

                        # Progress Update & Bookkeeping
                        bixel_counter += curr_ray["num_of_bixels"]
                        bixel_beam_counter += curr_ray["num_of_bixels"]

        # Finalize dose calculation
        logger.info("Finalizing dose calculation...")
        t_start = time.time()
        dij = self._finalize_dose(dij)
        t_end = time.time()
        logger.info("Done in %f seconds.", t_end - t_start)

        return dij

    def _init_dose_calc(self, ct: CT, cst: StructureSet, stf: SteeringInformation):
        """
        Initialize the dose calculation.

        Modified inherited method of the superclass DoseEngine,
        containing initialization which is specifically needed for
        pencil beam calculation and not for other engines.
        """

        dij = super()._init_dose_calc(ct, cst, stf)

        # calculate rED or rSP from HU or take provided wedCube
        if self.use_given_eq_density_cube and not hasattr(ct, "cube_hu"):
            logging.warning(
                "HU Conversion requested to be omitted but no ct.cube exists! "
                "Will override and do the conversion anyway!"
            )
            self.use_given_eq_density_cube = False

        if self.use_given_eq_density_cube:
            ct_wed = ct.cube_hu
            logging.info("Omitting HU to rED/rSP conversion and using existing ct.cube!\n")
        else:
            # TODO: obtain correct hlut
            if self.hlut is None:
                self.hlut = default_hlut()
            ct_wed = ct.compute_wet(self.hlut)

        self.cube_wed = ct_wed

        # Ignore densities outside of contours
        if self.ignore_outside_densities:  # TODO: default = None (not tested yet)
            mask_image = np2sitk.linear_indices_to_sitk_mask(
                self._vct_grid, self.cube_wed, order="numpy"
            )

            # # TODO: ct does not yet support multi_scen
            # for i in range(ct.num_of_ct_scen):
            #     self.cube_wed.ToScalarImage()
            #     self.cube_wed[erase_ct_dens_mask == 1] = 0

            # Apply the mask to the cube_wed image
            self.cube_wed = sitk.Mask(self.cube_wed, mask_image, outsideValue=0)

        # Allocate memory for quantity containers
        dij = self._allocate_quantity_matrices(dij, ["physical_dose"])

        # Initialize ray-tracer
        if self.trace_on_dose_grid:
            wed_cube_trace = resample_image(
                input_image=self.cube_wed,
                interpolator=sitk.sitkLinear,
                target_grid=self.dose_grid,
            )
        else:
            wed_cube_trace = self.cube_wed

        self._raytracer = RayTracerSiddon([wed_cube_trace])

        return dij

    def _allocate_quantity_matrices(self, dij: dict[str, Any], names: list[str]):
        # Loop over all requested quantities
        for q_name in names:
            # Create dij list for each quantity
            dij[q_name] = np.empty(self.mult_scen.scen_mask.shape, dtype=object)

            # Loop over all scenarios and preallocate quantity containers
            # TODO: write test for this
            for i in range(self.mult_scen.scen_mask.size):
                # Only if there is a scenario we will allocate
                if self.mult_scen.scen_mask.flat[i]:
                    if self._calc_dose_direct:
                        dij[q_name].flat[i] = np.zeros(
                            (self.dose_grid.num_voxels, dij["num_of_beams"]), dtype=np.float32
                        )
                    else:
                        # We allocate raw csc sparse matrix structures using
                        # a rough estimat ebased on number of voxels and
                        # beamlets
                        est_nnz = np.fix(
                            5e-4 * self.dose_grid.num_voxels * self._num_of_columns_dij
                        ).astype(np.int64)
                        dij[q_name].flat[i] = {
                            "data": np.empty((est_nnz,), dtype=np.float32),
                            "indices": np.empty((est_nnz,), dtype=np.int64),
                            "indptr": np.zeros((self._num_of_columns_dij + 1,), dtype=np.int64),
                            "nnz": 0,
                        }

            self._computed_quantities.append(q_name)

        self._effective_lateral_cutoff = self.geometric_lateral_cutoff

        return dij

    def _init_beam(
        self, _dij: dict, ct: CT, _cst: StructureSet, stf: SteeringInformation, i
    ) -> dict:
        """
        Initialize the beam for pencil beam dose calculation.

        Parameters
        ----------
        dij : dict
            The dose influence matrix dictionary.
        ct : CT
            The CT object.
        _cst : StructureSet
            The structure set object. Unused here
        stf : SteeringInformation
            The steering information object.
        i : int
            Index of the beam.

        Returns
        -------
        dict
            Beam Information dictionary
        """
        # TODO: here .model_dump() is used to get beam as dict. Its possible to change...
        # ... the ray_tracing to handle this as the model
        beam_info = {"beam": stf.beams[i].model_dump(), "beam_index": i}

        # Convert voxel indices to real coordinates using iso center of beam i
        coords_v = self._vox_world_coords - beam_info["beam"]["iso_center"]
        coords_vdose_grid = self._vox_world_coords_dose_grid - beam_info["beam"]["iso_center"]

        # Get Rotation Matrix
        beam_info["rot_mat_system_T"] = get_beam_rotation_matrix(
            beam_info["beam"]["gantry_angle"], beam_info["beam"]["couch_angle"]
        )

        # Rotate coordinates (1st couch around Y axis, 2nd gantry movement)
        rot_coords_v = np.dot(coords_v, beam_info["rot_mat_system_T"])
        rot_coords_vdose_grid = np.dot(coords_vdose_grid, beam_info["rot_mat_system_T"])

        rot_coords_v -= beam_info["beam"]["source_point_bev"]
        rot_coords_vdose_grid -= beam_info["beam"]["source_point_bev"]

        # Calculate geometric distances
        geo_dist_vdose_grid = [
            np.sqrt(np.sum(rot_coords_vdose_grid**2, axis=1)) for _ in range(ct.num_of_ct_scen)
        ]

        # Calculate radiological depth cube
        logger.info("Calculating radiological depth cube...")

        start_time = time.time()

        self._raytracer.debug_core_performance = True
        rad_depth_cubes = self._raytracer.trace_cubes(beam_info["beam"])
        self._raytracer.debug_core_performance = False

        # TODO: add universal time-debugging method
        end_time = time.time()
        elapsed_time = end_time - start_time
        logger.info("Elapsed time for rayTracing per Beam: %f seconds", elapsed_time)

        beam_info["valid_coords"] = [None] * len(rad_depth_cubes)
        beam_info["rad_depths"] = [None] * len(rad_depth_cubes)

        for c, rad_depth_cube in enumerate(rad_depth_cubes):
            if self.trace_on_dose_grid:
                rad_depth_cube_dose_grid = rad_depth_cube
            else:
                rad_depth_cube_dose_grid = resample_image(
                    input_image=rad_depth_cube,
                    interpolator=sitk.sitkLinear,
                    target_grid=self.dose_grid,
                )
            if self.keep_rad_depth_cubes:
                self._rad_depth_cubes.append(rad_depth_cube_dose_grid)

            rad_depth_cube_dosegrid = sitk.GetArrayViewFromImage(rad_depth_cube_dose_grid)
            rad_depth_vdose_grid = rad_depth_cube_dosegrid.ravel()[self._vdose_grid]

            # Find valid coordinates
            coord_is_valid = np.isfinite(rad_depth_vdose_grid)

            beam_info["valid_coords"][c] = coord_is_valid

            # TODO: !remove brakets once mutliscen is implemented
            beam_info["rad_depths"][c] = rad_depth_vdose_grid

        beam_info["geo_depths"] = geo_dist_vdose_grid
        beam_info["bev_coords"] = rot_coords_vdose_grid

        beam_info["valid_coords_all"] = np.any(np.vstack(beam_info["valid_coords"]), axis=0)

        # Check existence of target_points
        if any(r["target_point"] is None for r in beam_info["beam"]["rays"]):
            logger.debug("Missing target_point in rays. Calculating target points...")
            for ray in beam_info["beam"]["rays"]:
                ray["target_point"] = 2 * ray["ray_pos"] - beam_info["beam"]["source_point"]
                ray["target_point_bev"] = (
                    2 * ray["ray_pos_bev"] - beam_info["beam"]["source_point_bev"]
                )

        # Compute SSDs
        self._compute_ssd(beam_info, ct, density_threshold=self.ssd_density_threshold)

        logger.info("Done.")

        return beam_info

    def _init_ray(self, beam_info: dict[str], j: int) -> dict[str]:
        """
        Initialize a ray for pencil beam dose calculation.

        Parameters
        ----------
        curr_beam : dict
            The current beam data.
        j : int
            The ray index.

        Returns
        -------
        dict
            The initialized ray.
        """
        ray = beam_info["beam"]["rays"][j]
        ray["beam_index"] = beam_info["beam_index"]
        ray["ray_index"] = j
        ray["iso_center"] = beam_info["beam"]["iso_center"]

        if "num_of_bixels_per_ray" not in beam_info["beam"]:
            ray["num_of_bixels"] = 1
        else:
            ray["num_of_bixels"] = beam_info["beam"]["num_of_bixels_per_ray"][j]

        ray["source_point_bev"] = beam_info["beam"]["source_point_bev"]
        ray["sad"] = beam_info["beam"]["sad"]
        ray["bixel_width"] = beam_info["beam"]["bixel_width"]

        self._get_ray_geometry_from_beam(ray, beam_info)

        return ray

    def _compute_ssd(
        self,
        beam_info: dict,
        _ct: CT,
        mode: Literal["first"] = "first",
        density_threshold: float = 0.05,
        show_warning: bool = True,
    ):
        """
        Compute SSD (Source to Surface Distance) for each ray.

        Parameters
        ----------
        beam_info : dict
            Beam Information, will be modified to include SSD.
        ct : CT
            The CT object.
        mode : Literal["first"], optional
            Mode for handling multiple cubes to compute one SSD. Only 'first' is implemented.
        density_threshold : float, optional
            Value determining the skin threshold.
        show_warning : bool, optional
            Flag to show warnings.
        """

        beam = beam_info["beam"]

        # TODO: Add MultiScenario Support - remove this line
        if mode == "first":
            ssd = [None] * beam["num_of_rays"]

            ray_pos_bev = np.array([ray["ray_pos_bev"] for ray in beam["rays"]])
            target_points = np.array([ray["target_point"] for ray in beam["rays"]])

            alpha, _, rho, d12, _ = self._raytracer.trace_rays(
                beam["iso_center"],
                beam["source_point"].reshape((1, 3)),
                target_points,
            )

            # find rays that do not hit patients
            ix_nan = np.all(np.isnan(rho[0]), axis=1)
            if np.any(ix_nan):
                msg = f"{ix_nan.sum()} rays do not hit patient. Trying to fix afterwards..."
                if show_warning:
                    warnings.warn(msg)
                else:
                    logger.warning(msg)

            ix_ssd = -1 * np.ones_like(ix_nan, dtype=np.int64)
            ix_ssd[~ix_nan] = np.argmax(rho[0][~ix_nan] > density_threshold, axis=1)

            ssd = beam["sad"] * np.ones_like(ix_ssd, dtype=np.float32)

            ssd[~ix_nan] = d12[~ix_nan].squeeze() * alpha[~ix_nan, ix_ssd[~ix_nan]]

            # Now assign the ssd to all rays
            for j, ray in enumerate(beam["rays"]):
                if ix_nan[j]:  # Try to fix SSD by using SSD of closest neighboring ray
                    ray["SSD"] = self._closest_neighbor_ssd(ray_pos_bev, ssd, ray["ray_pos_bev"])
                else:
                    ray["SSD"] = float(ssd[j])

        else:
            raise ValueError(f"Invalid mode {mode} for SSD calculation")

        beam_info["beam"] = beam

    def _closest_neighbor_ssd(
        self, ray_pos_bev: np.ndarray, ssd: np.ndarray, curr_pos: np.ndarray
    ) -> float:
        """
        Find the closest neighboring ray's SSD.

        Parameters
        ----------
        ray_pos_bev : np.ndarray
            Array of ray positions.
        curr_pos : np.ndarray
            Current ray position.

        Returns
        -------
        float
            SSD value of the closest neighboring ray.
        """
        distances = np.sum((ray_pos_bev - curr_pos) ** 2, axis=1)
        sorted_indices = np.argsort(distances)
        for ix in sorted_indices:
            if ssd[ix] is not None:
                return float(ssd[ix])
        raise ValueError(
            "Error in SSD calculation: Could not fix SSD calculation by using closest neighboring "
            "ray."
        )

        #
        # """Find the closest neighbor to ray j and return its SSD."""
        # distances = np.linalg.norm(ray_pos_bev - ray_pos_bev[j], axis=1)
        # distances[j] = np.inf
        # return SSD[np.argmin(distances)]

    def _extract_single_scenario_ray(self, ray: dict, scen_idx: int):
        """
        Extract a single scenario ray and adapt radiological depths.

        Parameters
        ----------
        ray (dict):
            The ray data.
        scen_idx (int):
            The scenario index.

        Returns
        -------
        dict:
            The scenario ray with adapted radiological depths.
        """
        # Gets number of scenario
        scen_num = self.mult_scen.scen_num(scen_idx)
        ct_scen = self.mult_scen.linear_mask[0][scen_num]

        # First, create a ray of the specific scenario to adapt rad depths
        scen_ray = ray.copy()
        scen_ray["rad_depths"] = scen_ray["rad_depths"][ct_scen]
        scen_ray["rad_depths"] = (1 + self.mult_scen.rel_range_shift[scen_num]) * scen_ray[
            "rad_depths"
        ] + self.mult_scen.abs_range_shift[scen_num]
        scen_ray["radial_dist_sq"] = scen_ray["radial_dist_sq"][ct_scen]
        scen_ray["ix"] = scen_ray["ix"][ct_scen]

        if self.mult_scen.abs_range_shift[scen_num] < 0:
            # TODO: better way to handle this?
            scen_ray["rad_depths"][scen_ray["rad_depths"] < 0] = 0

        if "geo_depths" in scen_ray:
            scen_ray["geo_depths"] = scen_ray["geo_depths"][ct_scen]

        if "lat_dists" in scen_ray:
            scen_ray["lat_dists"] = scen_ray["lat_dists"][ct_scen]

        if "iso_lat_dists" in scen_ray:
            scen_ray["iso_lat_dists"] = scen_ray["iso_lat_dists"][ct_scen]

        return scen_ray

    def _get_ray_geometry_from_beam(self, ray: dict[str], beam_info: dict[str]):
        ray["effective_lateral_cut_off"] = beam_info.get(
            "effective_lateral_cut_off", self._effective_lateral_cutoff
        )
        lateral_ray_cutoff = self._get_lateral_distance_from_dose_cutoff_on_ray(ray)

        # Ray tracing for beam i and ray j
        ix, radial_dist_sq, lat_dists, iso_lat_dists = self.calc_geo_dists(
            beam_info["bev_coords"],
            ray["source_point_bev"],
            ray["target_point_bev"],
            ray["sad"],
            beam_info["valid_coords_all"],
            lateral_ray_cutoff,
        )

        # Subindex given the relevant indices from the geometric distance calculation
        ray["valid_coords"] = [beam_ix & ix for beam_ix in beam_info["valid_coords"]]
        ray["ix"] = [self._vdose_grid[ix_in_grid] for ix_in_grid in ray["valid_coords"]]

        ray["radial_dist_sq"] = [radial_dist_sq[beam_ix[ix]] for beam_ix in ray["valid_coords"]]
        ray["lat_dists"] = [lat_dists[beam_ix[ix]] for beam_ix in ray["valid_coords"]]
        ray["iso_lat_dists"] = [iso_lat_dists[beam_ix[ix]] for beam_ix in ray["valid_coords"]]

        ray["valid_coords_all"] = np.any(np.vstack(ray["valid_coords"]), axis=0)

        ray["geo_depths"] = [
            rD[ix] for rD, ix in zip(beam_info["geo_depths"], ray["valid_coords"])
        ]  # usually not needed for particle beams
        ray["rad_depths"] = [
            rD[ix] for rD, ix in zip(beam_info["rad_depths"], ray["valid_coords"])
        ]

    def _get_lateral_distance_from_dose_cutoff_on_ray(self, ray: dict) -> float:
        """
        Obtain the maximum lateral cutoff on a a ray.

        Distance will computed from dosimetric cutoff setting.

        Parameters
        ----------
        _ray : dict
            The ray data. Unused in this base implementation.

        Returns
        -------
        float
            The lateral distance from the dose cutoff on the ray.
        """

        return ray.get("effective_lateral_cut_off", self._effective_lateral_cutoff)

    def _fill_dij(
        self,
        bixel: dict,
        dij: dict,
        _stf: SteeringInformation,
        scen_idx: int,
        curr_beam_idx: int,
        curr_ray_idx: int,
        curr_bixel_idx: int,
        bixel_counter: int,
    ):
        """
        Fill the dose influence matrix (dij) with bixel contents.

        This is the last step in bixel dose calculation. It will fill all
        the computed quantities into sparse matrix containers.
        If forward calculation is active, accumulation into dense vectors
        will be performed instead.

        Parameters
        ----------
        bixel : dict
            The bixel data.
        dij : dict
            The dose influence matrix.
        stf : SteeringInformation
            The structure containing beam information.
            Unused in this base implementation.
        scen_idx : int
            The scenario index.
        curr_beam_idx : int
            The current beam index.
        curr_ray_idx : int
            The current ray index.
        curr_bixel_idx : int
            The current bixel index.
        bixel_counter : int
            The counter for the bixels.

        Returns
        -------
        dict
            The updated dose influence matrix.
        """
        # Only fill if we actually had bixel (indices) to compute
        if bixel and "ix" in bixel and bixel["ix"].any():
            sub_scen_idx = [
                np.unravel_index(scen_idx, self.mult_scen.scen_mask.shape)[i]
                for i in range(self.mult_scen.scen_mask.ndim)
            ]
            sub_scen_idx = tuple(sub_scen_idx)

            for q_name in self._computed_quantities:
                if self._calc_dose_direct:
                    # We accumulate the current bixel to the container
                    dij[q_name][sub_scen_idx][bixel["ix"], curr_beam_idx] += (
                        bixel["weight"] * bixel[q_name]
                    )
                else:
                    # first we fill the index pointer array
                    dij[q_name][sub_scen_idx]["indptr"][bixel_counter + 1] = (
                        dij[q_name][sub_scen_idx]["indptr"][bixel_counter] + bixel["ix"].size
                    )

                    # If we haven't preallocated enough, we need to expand the arrays
                    if (
                        dij[q_name][sub_scen_idx]["data"].size
                        < dij[q_name][sub_scen_idx]["nnz"] + bixel["ix"].size
                    ):
                        logger.debug("Resizing data and indices arrays for %s...", q_name)

                        # We estimate the required size by the remaining
                        # number of beamlets / columns in the sparse matrix
                        # 1 additional beamlet is added
                        shape_resize = (self._num_of_columns_dij - bixel_counter) * (
                            bixel["ix"].size + 1
                        )
                        shape_resize += dij[q_name][sub_scen_idx]["data"].size
                        dij[q_name][sub_scen_idx]["data"].resize((shape_resize,), refcheck=False)
                        dij[q_name][sub_scen_idx]["indices"].resize(
                            (shape_resize,), refcheck=False
                        )

                    # Fill the corresponding values and indices using indptr
                    dij[q_name][sub_scen_idx]["data"][
                        dij[q_name][sub_scen_idx]["indptr"][bixel_counter] : dij[q_name][
                            sub_scen_idx
                        ]["indptr"][bixel_counter + 1]
                    ] = bixel[q_name]
                    dij[q_name][sub_scen_idx]["indices"][
                        dij[q_name][sub_scen_idx]["indptr"][bixel_counter] : dij[q_name][
                            sub_scen_idx
                        ]["indptr"][bixel_counter + 1]
                    ] = bixel["ix"]

                    # Store how many nnzs we actually have in the matrix
                    dij[q_name][sub_scen_idx]["nnz"] += bixel["ix"].size

        # Bookkeeping of bixel numbers
        # remember beam and bixel number
        if self._calc_dose_direct:
            dij["beam_num"][curr_beam_idx] = curr_beam_idx
            dij["ray_num"][curr_beam_idx] = curr_beam_idx
            dij["bixel_num"][curr_beam_idx] = curr_beam_idx
        else:
            dij["beam_num"][bixel_counter] = curr_beam_idx
            dij["ray_num"][bixel_counter] = curr_ray_idx
            dij["bixel_num"][bixel_counter] = curr_bixel_idx

    def _finalize_dose(self, dij: dict):
        """
        Finalize the dose influence matrix.

        Pruning the matrix and concatenating the containers to a compressed
        sparse matrix.

        Parameters
        ----------
        dij : dict
            The dose influence matrix.

        Returns
        -------
        Dij
            The finalized dose influence matrix.
        """

        # Loop over all scenarios and remove dose influence for voxels outside of segmentations
        for i in range(self.mult_scen.scen_mask.size):
            # Only if there is a scenario we will allocate
            if self.mult_scen.scen_mask.flat[i]:
                # Loop over all used quantities
                for q_name in self._computed_quantities:
                    if not self._calc_dose_direct:
                        # tmp_matrix = cast(sparse.lil_matrix, dij[q_name].flat[i])
                        # tmp_matrix = tmp_matrix.tocsr().T
                        data_dict = dij[q_name].flat[i]

                        # Resize to the actual number of non-zero elements
                        data_dict["data"].resize((data_dict["nnz"],), refcheck=False)
                        data_dict["indices"].resize((data_dict["nnz"],), refcheck=False)

                        # Create the matrix, avoid copies
                        tmp_matrix = sparse.csc_array(
                            (data_dict["data"], data_dict["indices"], data_dict["indptr"]),
                            shape=(self.dose_grid.num_voxels, self._num_of_columns_dij),
                            dtype=np.float32,
                            copy=False,
                        )

                        # Do we need this?
                        tmp_matrix.eliminate_zeros()

                        # make sure indices are sorted and matrix is canonical
                        if not tmp_matrix.has_sorted_indices:
                            logger.debug("Sorting indices for %s...", q_name)
                            tmp_matrix.sort_indices()

                        if not tmp_matrix.has_canonical_format:
                            logger.debug("Matrix is not in canonical format for %s...", q_name)
                            tmp_matrix.sum_duplicates()

                        dij[q_name].flat[i] = tmp_matrix

        if self.keep_rad_depth_cubes and self._rad_depth_cubes:
            dij["rad_depth_cubes"] = self._rad_depth_cubes

        # Call the finalizeDose method from the base class
        return super()._finalize_dose(dij)

    @staticmethod
    def calc_geo_dists(
        rot_coords_bev, source_point_bev, target_point_bev, sad, rad_depth_ix, lateral_cutoff
    ):
        """
        Calculate geometric distances for dose calculation.

        Parameters
        ----------
        rot_coords_bev : ndarray
            Coordinates in beam's eye view (BEV) of the voxels where ray tracing results are
            available.
        source_point_bev : ndarray
            Source point in voxel coordinates in BEV.
        target_point_bev : ndarray
            Target point in voxel coordinates in BEV.
        sad : float
            Source-to-axis distance.
        rad_depth_ix : ndarray
            Subset of voxels for which radiological depth calculations are available.
        lateral_cutoff : float
            Lateral cutoff specifying the neighborhood for dose calculations.

        Returns
        -------
        ix : ndarray
            Indices of voxels where dose influence is computed.
        rad_distances_sq : ndarray
            Squared radial distances to the central ray.
        lat_dists : ndarray
            Lateral distances to the central ray (in X & Z).
        iso_lat_dists : ndarray
            Lateral distances to the central ray projected onto the isocenter plane.
        rot_coords_bev, source_point_bev, target_point_bev, sad, rad_depth_ix, lateral_cutoff
        """
        # Put [0 0 0] position in the source point for beamlet who passes through isocenter
        a = -source_point_bev.T

        # Normalize the vector
        a = a / np.linalg.norm(a)

        # Put [0 0 0] position in the source point for a single beamlet
        b = (target_point_bev - source_point_bev).T

        # Normalize the vector
        b = b / np.linalg.norm(b)

        # Define rotation matrix
        if np.all(a == b):
            rot_coords_temp = rot_coords_bev[rad_depth_ix, :]
        else:

            def ssc(v: np.ndarray) -> np.ndarray:
                """Skew-symmetric cross product matrix."""
                return np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])

            derived_rot_mat = (
                np.eye(3)
                + ssc(np.cross(a, b))
                + np.dot(ssc(np.cross(a, b)), ssc(np.cross(a, b)))
                * (1 - np.dot(a, b))
                / (np.linalg.norm(np.cross(a, b)) ** 2)
            )
            rot_coords_temp = np.dot(rot_coords_bev[rad_depth_ix, :], derived_rot_mat)

        # Put [0 0 0] position CT in center of the beamlet
        lat_dists = rot_coords_temp[:, [0, 2]] + source_point_bev[[0, 2]]

        # Check if radial distance exceeds lateral cutoff (projected to iso center)
        rad_distances_sq = np.sum(lat_dists**2, axis=1)
        subset_mask = rad_distances_sq <= (lateral_cutoff / sad) ** 2 * rot_coords_temp[:, 1] ** 2

        ix = rad_depth_ix.copy()
        # Apply mask for return quantities
        ix[rad_depth_ix] = subset_mask

        # Return radial distances squared
        rad_distances_sq = rad_distances_sq[subset_mask]

        # Lateral distances in X & Z
        lat_dists = lat_dists[subset_mask, :]

        # Lateral distances projected onto isocenter
        iso_lat_dists = lat_dists / rot_coords_temp[subset_mask, 1][:, np.newaxis] * sad

        return ix, rad_distances_sq, lat_dists, iso_lat_dists



"""
The :mod: 'pyanno4rt.optimization' module implements all classes related to
dose optimization, including objective and constraint functions, solvers and
backprojection classes.
"""

# Author: Tim Ortkamp <tim.ortkamp@kit.edu>, 2023

from .objectives import Objective
from ._fluence_optimization import fluence_optimization

__all__ = ["fluence_optimization", "Objective"]



import numpy as np

from pyRadPlan.ct import CT, validate_ct
from pyRadPlan.cst import StructureSet, validate_cst
from pyRadPlan.plan import Plan, validate_pln
from pyRadPlan.dij import Dij, validate_dij
from pyRadPlan.stf import SteeringInformation, validate_stf

from .problems import get_problem_from_pln


def fluence_optimization(
    ct: CT, cst: StructureSet, stf: SteeringInformation, dij: Dij, pln: Plan
) -> np.ndarray:
    """
    Trigger fluence optimization using the configuration stored in the pln object.

    Parameters
    ----------
    ct : CT
        CT object.
    cst : StructureSet
        StructureSet object.
    stf : SteeringInformation
        SteeringInformation object.
    dij : Dij
        Dij object.
    pln : Plan
        Plan object.

    Returns
    -------
    np.ndarray
        The optimized fluence map.
    """

    _ct = validate_ct(ct)
    _cst = validate_cst(cst)
    _stf = validate_stf(stf)
    _dij = validate_dij(dij)
    _pln = validate_pln(pln)

    planning_prob = get_problem_from_pln(_pln)

    x, _result_info = planning_prob.solve(_ct, _cst, _stf, _dij)

    return x



from typing import Optional
import os
import warnings
import logging
import sys

if sys.version_info < (3, 10):
    import importlib_resources as resources  # Backport for older versions
else:
    from importlib import resources  # Standard from Python 3.9+

from pyRadPlan.ct import validate_ct, CT
from pyRadPlan.cst import validate_cst, StructureSet
from pyRadPlan.stf import validate_stf
from pyRadPlan.plan import validate_pln
from pyRadPlan.dij import validate_dij
from . import matfile

logger = logging.getLogger(__name__)


def load_tg119() -> tuple[CT, StructureSet]:
    """
    Load the included TG119 phantom.

    This is a helper function to load the TG119 phantom included in the
    package data using importlib.resources.

    Returns
    -------
    tuple[CT, StructureSet]
        The CT and StructureSet objects.
    """
    phantom_data_str = resources.files("pyRadPlan.data.phantoms").joinpath("TG119.mat")
    return load_patient(phantom_data_str)


def load_patient(
    filename: os.PathLike,
    extra_plan_data: Optional[dict] = None,
    extra_data: Optional[dict] = None,
) -> tuple[CT, StructureSet]:
    """
    Load a patient from a file.

    Chooses loader from the extension.

    Parameters
    ----------
    filename : os.PathLike
        Path to the file.
    additional_plan_data : Optional[dict]
        Additional data structures known to pyRadPlan that were present in the mat file
    """

    # Sanitize path and check if file exists
    path = os.path.normpath(filename)
    if not os.path.exists(path):
        raise FileNotFoundError(f"Patient file not found: {path}")

    # Get the file extension
    ext = os.path.splitext(path)[1].lower()

    # Load the patient
    if ext == ".mat":
        mdict = matfile.load(path)
        patient_dict = validate_matrad_patient(mdict)
    else:
        raise ValueError(f"Unsupported file extension: {ext}")

    # we require ct
    ct = patient_dict.pop("ct", None)
    if ct is None:
        raise ValueError("ct is missing from the patient file.")

    # we issue a warning when getting the cst
    cst = patient_dict.pop("cst", None)
    if cst is None:
        warnings.warn("cst/StructureSet is missing from the patient file.")

    extra_plan_data.update(patient_dict) if isinstance(extra_plan_data, dict) else None
    extra_data.update(mdict) if isinstance(extra_data, dict) else None

    return ct, cst


def validate_matrad_patient(mdict: dict[str], remove_matrad_structures: bool = True) -> dict[str]:
    """
    Load a matRad-like patient from a mat file.

    Assumes that the file uses matRad's data structures and tries to validates them.

    Parameters
    ----------
    mdict : dict[str]
        Dictionary imported from a .mat file. Modified if remove_matrad_structures is True.
    remove_matrad_structures : bool
        Pop the input data recognized/named as matRad structures from the dictionary.

    Returns
    -------
    dict[str]
        A dictionary with the validated data.

    """

    patient_dict = {}

    ct = mdict.pop("ct", None)
    cst = mdict.pop("cst", None)

    if ct is not None:
        patient_dict["ct"] = validate_ct(ct)

    if cst is not None:
        patient_dict["cst"] = validate_cst(cst, patient_dict["ct"])

    for key, validator in [("pln", validate_pln), ("stf", validate_stf), ("dij", validate_dij)]:
        value = mdict.get(key, None)
        if value is not None:
            try:
                patient_dict[key] = validator(value)
                if remove_matrad_structures:
                    mdict.pop(key)
            except ValueError:
                logger.warning(f"{key} present but could not be validated.")

    result = mdict.get("resultGUI", None)
    if result is not None:
        # TODO: validation as soon as result structure is implemented
        patient_dict["result"] = result
        if remove_matrad_structures:
            mdict.pop("resultGUI")

    return patient_dict



import logging
from typing import Union, Optional, Any
from typing_extensions import Self
from pydantic import (
    Field,
    computed_field,
    field_validator,
    model_validator,
)
from numpydantic import NDArray, Shape

import pint
import numpy as np
import SimpleITK as sitk
import matplotlib.pyplot as plt

from pyRadPlan.core import PyRadPlanBaseModel
from pyRadPlan.cst import StructureSet

logger = logging.getLogger(__name__)

ureg = pint.UnitRegistry()


class DVH(PyRadPlanBaseModel):
    """
    Dose Volume Histogram (DVH) class.

    Note
    ----
    The DVH class can be used for other quantities as well.
    The bins correspond to the left edge of the bin, thus we have as many bin points as dose points.

    """

    diff_volume: NDArray[Shape["*"], np.float64] = Field(
        alias="volumePoints", description="Volume percentages for each point"
    )

    unit: pint.Unit = Field(default=ureg.gray, description="Unit of the DVH bins")
    name: str = Field(default="DVH", description="Name of the DVH")

    bin_edges: NDArray[Shape["*"], np.float64] = Field(
        alias="doseGrid", description="Left bin edges for the DVH"
    )

    @field_validator("unit", mode="before")
    @classmethod
    def validate_unit(cls, v: Any) -> pint.Unit:
        """
        Validate the unit of the DVH bins.

        Parameters
        ----------
        v : Any
            The unit to validate.

        Returns
        -------
            pint.Unit: Validated unit of the DVH bins.
        """

        if isinstance(v, pint.Unit):
            return v

        try:
            return ureg(v)
        except pint.UndefinedUnitError:
            raise ValueError(f"Invalid unit: {v}")

    @model_validator(mode="after")
    def check_shapes(self) -> Self:
        """
        Validate the shapes of the DVH data.

        Returns
        -------
            DVH: The DVH object itself if validation passes.
        """
        if (
            self.diff_volume.ndim != self.bin_edges.ndim
            or self.diff_volume.ndim != 1
            or self.bin_edges.size != self.diff_volume.size + 1
        ):
            raise ValueError("The shape of diff_volume must be consistent with shape of bins.")
        return self

    @computed_field
    @property
    def cum_volume(self) -> NDArray[Shape["*"], np.float64]:
        """
        Get the cumulative DVH Volume.

        Returns
        -------
            NDArray: Cumulative DVH volume as a numpy array
        """
        return np.cumsum(self.diff_volume[::-1])[::-1]

    @computed_field
    @property
    def cumulative(self) -> NDArray[Shape["2,*"], np.float64]:
        """
        Get the cumulative DVH.

        Gives a 2xnum_points array with the first row being the dose / quantity
        and the second row being the cumulative volume percentages.

        Returns
        -------
            NDArray: Cumulative DVH

        Notes
        -----
        The x coordinate (dose/quantity) corresponds to the histograms left
        bin edge, as this corresponds to the meaning: Volume covered by
        "at least" the respective dose/quantity value.
        """
        return np.vstack((self.bins, self.cum_volume))

    @computed_field
    @property
    def differential(self) -> NDArray[Shape["2,*"], np.float64]:
        """
        Get the differential DVH.

        Gives a 2xnum_points array with the first row being the dose
        and the second row being the differential volume percentages.

        Returns
        -------
            NDArray: Differential DVH

        Notes
        -----
        The x ccoordinate (dose/quantity) corresponds to the histograms
        bin center for the differntial DVH. This is because the differential
        DVH captures the relative volume within a bin, thus the center
        represents the dose/quantity value for that bin best.
        Be aware of this difference to the cumulative DVH.

        """
        return np.vstack((self.bin_centers, self.diff_volume))

    @computed_field
    @property
    def num_points(self) -> int:
        """
        Get the number of points in the DVH.

        Returns
        -------
            int: Number of points in the DVH
        """
        return len(self.diff_volume)

    @computed_field
    @property
    def bins(self) -> NDArray[Shape["*"], np.float64]:
        """
        Get the DVH bin left edges.

        This corresponds to the dose / quantity thresholds for the cumulative volume percentages.

        Returns
        -------
            NDArray: DVH bins as a numpy array
        """
        return self.bin_edges[:-1]

    @computed_field
    @property
    def has_regular_bins(self) -> bool:
        """
        Check if the DVH has regular bins.

        Returns
        -------
            bool: True if the DVH has regular bins, False otherwise
        """
        return np.all(np.diff(self.bin_edges) == self.bin_edges[1] - self.bin_edges[0])

    @computed_field
    @property
    def bin_centers(self) -> NDArray[Shape["*"], np.float64]:
        """
        Get the centers of the DVH bins.

        Returns
        -------
            NDArray: Centers of the DVH bins as a numpy array
        """
        return (self.bin_edges[:-1] + self.bin_edges[1:]) / 2.0

    def get_vx(self, x: Union[float, NDArray]) -> float:
        """
        Get the volume percentage for a given dose (or other quantity) value.

        Parameters
        ----------
        x : float
            The dose/quantity value to get the volume percentage for.

        Returns
        -------
            float: Volume percentage corresponding to the given dose/quantity value.
        """
        if x < self.bins[0] or x > self.bins[-1]:
            raise ValueError("Dose value is out of bounds of the DVH bins.")

        # Interpolate to find the volume percentage
        return np.interp(x, self.bins, self.cum_volume)

    def get_dy(self, y: Union[float, NDArray]) -> float:
        """
        Get the dose (or other quantity) value for a given volume percentage.

        Parameters
        ----------
        y : float
            The volume percentage to get the dose/quantity value for.

        Returns
        -------
            float: Dose/quantity value corresponding to the given volume percentage.
        """
        if y < 0 or y > 100:
            raise ValueError("Volume percentage must be between 0 and 100.")

        # Interpolate to find the dose value
        return np.interp(y / 100, self.cum_volume, self.bins)

    @classmethod
    def compute(
        cls,
        quantity: Union[sitk.Image, NDArray],
        mask: Optional[Union[sitk.Image, NDArray]] = None,
        num_points: int = 1000,
        max_value: Optional[float] = None,
        min_value: float = 0.0,
        **kwargs,
    ) -> Self:
        """
        Create a DVH from a dose_like quantity and mask.

        Parameters
        ----------
        quantity : Union[sitk.Image, NDArray]
            The dose-like quantity to update the DVH with. Can be a SimpleITK image or a numpy array.
        mask : Optional[Union[sitk.Image, NDArray]], optional
            The mask to apply to the dose quantity. If None, the whole image / array will be used.
            If provided, it must either have the same shape as the quantity.
            Defaults to None.
        num_points : int, optional
            The number of points in the DVH. Defaults to 100.
        max_value : Optional[float], optional
            The maximum value for the DVH bins. If None, it will be set to 105% of the maximum value in the quantity.
        min_value : float, optional
            The minimum value for the DVH bins. Defaults to 0.0.
        **kwargs: Additional arguments passed to the DVH Model.

        Returns
        -------
            DVH: DVH Object
        """

        # manage quantity
        if isinstance(quantity, sitk.Image):
            # Convert SimpleITK image to numpy array
            quantity = sitk.GetArrayFromImage(quantity)

        # First check if mask is provided and is valid
        if mask is not None:
            if isinstance(mask, sitk.Image):
                # Convert SimpleITK mask to numpy array
                mask = sitk.GetArrayViewFromImage(mask)

            # Mask has either same dimensionality as quantity, or is a linear mask
            if mask.shape != quantity.shape:
                raise ValueError("Mask must have the same shape as the quantity.")

            q_array = quantity[mask.astype(np.bool)].flat
        else:
            q_array = quantity.flat

        if max_value is None:
            max_value = np.max(q_array) * 1.05

        # Use bin edges that include the last point
        hist, edges = np.histogram(
            q_array, bins=num_points, range=(min_value, max_value), density=False
        )
        # Remove the last bin
        hist = hist / len(q_array) * 100

        return DVH(diff_volume=hist, bin_edges=edges, **kwargs)

    def plot(self, ax=None, line_width=2, plot_legend=True, **kwargs):
        """Plot the DVH curve.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. If None, creates new figure and axes. Defaults to None.
        line_width : float, optional
            Width of the plotted lines. Defaults to 2.
        plot_legend : bool, optional
            Whether to show the legend. Defaults to True.
        **kwargs:
            Additional arguments passed to matplotlib's plot function.

        Returns
        -------
            matplotlib.axes.Axes: The axes containing the plot
        """
        if ax is None:
            _, ax = plt.subplots()

        # Plot with or without legend
        if plot_legend:
            ax.plot(
                self.bins,
                self.cum_volume,
                linewidth=line_width,
                label=self.name,
                **kwargs,
            )
        else:
            ax.plot(self.bins, self.cum_volume, linewidth=line_width, **kwargs)

        # Set labels and legend
        ax.set_xlabel("Dose [Gy]")
        ax.set_ylabel("Volume [%]")
        if plot_legend:
            ax.legend()

        # Set axis limits with 5% padding
        ax.set_xlim(0, np.max(self.bins) * 1.05)
        ax.set_ylim(0, np.max(self.cumulative) * 1.05)

        return ax


class DVHCollection(PyRadPlanBaseModel):
    """Collection of DVH curves for multiple structures."""

    dvhs: list[DVH] = Field(description="List of DVH curves")

    @classmethod
    def from_structure_set(
        cls,
        cst: StructureSet,
        dose: Union[NDArray, sitk.Image],
    ) -> Self:
        """
        Create DVH collection from a structure set.

        Parameters
        ----------
        cst : StructureSet
            The structure set containing the VOIs (Volumes of Interest).
        dose : Union[NDArray, sitk.Image]
            The dose-like quantity to compute the DVH from. Can be a SimpleITK image or a numpy array.

        Returns
        -------
            DVHCollection: Collection of DVH curves for all structures
        """
        dvhs = [DVH.compute(quantity=dose, mask=voi.mask, name=voi.name) for voi in cst.vois]
        return cls(dvhs=dvhs)

    def plot(
        self,
        structures: Union[list[str], None] = None,
        ax=None,
        line_width=2,
        plot_legend=True,
        **kwargs,
    ):
        """Plot DVH curves for selected or all structures.

        Parameters
        ----------
        structures : Union[list[str], None], optional
            List of structure names to plot. If None, all structures are plotted. Defaults to None.
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. If None, creates new figure and axes. Defaults to None
        line_width : float, optional
            Width of the plotted lines. Defaults to 2.
        plot_legend : bool, optional
            Whether to show the legend. Defaults to True.
        **kwargs:
            Additional arguments passed to matplotlib's plot function.


        Returns
        -------
            matplotlib.axes.Axes: The axes containing the plot
        """
        if ax is None:
            _, ax = plt.subplots()

        # Filter DVHs based on structure names if provided
        dvhs_to_plot = self.dvhs
        if structures is not None:
            dvhs_to_plot = [dvh for dvh in self.dvhs if dvh.name in structures]
            if not dvhs_to_plot:
                raise ValueError("None of the specified structures found in DVH collection")

        # Plot each DVH
        for dvh in dvhs_to_plot:
            dvh.plot(ax=ax, line_width=line_width, plot_legend=plot_legend, **kwargs)

        return ax



from typing import Optional, Union, Literal

import numpy as np
import SimpleITK as sitk
import matplotlib.pyplot as plt
import pint
from pyRadPlan import CT, validate_ct, StructureSet, validate_cst

# Initialize Units
ureg = pint.UnitRegistry()


def plot_slice(  # noqa: PLR0913
    ct: Optional[Union[CT, dict]] = None,
    cst: Optional[Union[StructureSet, dict, list]] = None,
    overlay: Optional[Union[sitk.Image, np.ndarray]] = None,
    view_slice: Optional[Union[list[int], np.ndarray, int]] = None,
    plane: Union[Literal["axial", "coronal", "sagittal"], int] = "axial",
    overlay_alpha: float = 0.5,
    overlay_unit: Union[str, pint.Unit] = pint.Unit(""),
    overlay_rel_threshold: float = 0.01,
    contour_line_width: float = 1.0,
    save_filename: Optional[str] = None,
    show_plot: bool = True,
    use_global_max: bool = False,
):
    """Plot one or multiple slices of the CT with a given overlay.

    Parameters
    ----------
    ct : CT
        The CT object.
    cst : StructureSet
        The StructureSet object.
    overlay : sitk.Image or np.ndarray
        The overlay image to visualize.
    view_slice : List[int], array, int or None
        Slice indices to visualize.
    plane : str or int
        The plane to visualize. Can be "axial", "coronal", or "sagittal".
    overlay_alpha : float
        The alpha value for the overlay.
    overlay_unit : str or pint.Unit
        The unit of the overlay.
    overlay_rel_threshold : float
        The relative threshold for the overlay.
    contour_line_width : float
        The line width for the contour lines.
    save_filename : str
        The filename to save the plot. Default is None
    show_plot : bool
        If True, show the plot. Default is True.
    use_global_max : bool
        If True, use the overlay's global maximum for sclaing
    """

    if ct is not None:
        ct = validate_ct(ct)
        cube_hu = sitk.GetArrayViewFromImage(ct.cube_hu)
        array_shape = cube_hu.shape

    if cst is not None:
        cst = validate_cst(cst)
        array_shape = cst.ct_image.size[::-1]

    if ct is None and cst is None:
        raise ValueError("Nothing to visualize!")

    plane = {"axial": 0, "coronal": 1, "sagittal": 2}.get(plane, plane)
    if not isinstance(plane, int) or not 0 <= plane <= 2:
        raise ValueError("Invalid plane")

    if isinstance(overlay_unit, str):
        overlay_unit = ureg(overlay_unit)

    if view_slice is None:
        view_slice = [int(np.round(array_shape[plane] / 2))]
    elif isinstance(view_slice, int):
        view_slice = [view_slice]

    num_slices = len(view_slice)
    cols = int(np.ceil(np.sqrt(num_slices)))
    rows = int(np.ceil(num_slices / cols))

    fig, axes = plt.subplots(rows, cols, figsize=(5 * cols, 5 * rows))
    axes = np.array(axes).flatten()

    # Prepare overlay if provided.
    if overlay is not None:
        if isinstance(overlay, sitk.Image):
            overlay = sitk.GetArrayViewFromImage(overlay)
        if use_global_max:
            global_overlay_max = np.max(overlay)

    for i, slice_idx in enumerate(view_slice):
        ax = axes[i]
        slice_indexing = tuple(slice(None) if j != plane else slice_idx for j in range(3))

        ax.tick_params(
            axis="both",
            which="both",
            bottom=False,
            top=False,
            labelbottom=False,
            left=False,
            right=False,
            labelleft=False,
        )

        # Visualize the CT slice.
        if ct is not None:
            ax.imshow(cube_hu[slice_indexing], cmap="gray")

        # Visualize the VOIs from the StructureSet.
        if cst is not None:
            for v, voi in enumerate(cst.vois):
                mask = sitk.GetArrayViewFromImage(voi.mask)
                cmap = plt.colormaps["cool"]
                color = cmap(v / len(cst.vois))
                ax.contour(
                    mask[slice_indexing],
                    levels=[0.5],
                    colors=[color],
                    linewidths=contour_line_width,
                )

        # Visualize the overlay.
        if overlay is not None:
            if use_global_max:
                current_max = global_overlay_max
            else:
                current_max = np.max(overlay[slice_indexing])
            im_overlay = ax.imshow(
                overlay[slice_indexing],
                cmap="jet",
                interpolation="nearest",
                alpha=overlay_alpha
                * (overlay[slice_indexing] > overlay_rel_threshold * current_max),
                vmin=0,
                vmax=current_max,
            )
            plt.colorbar(im_overlay, ax=ax, label=f"{overlay_unit:~P}")

        ax.set_title(f"Slice z={slice_idx}")

        # Add a scale bar if a CT is provided.
        if ct is not None:
            disp_im = cube_hu[slice_indexing]

            spacing = ct.cube_hu.GetSpacing()
            if plane in (0, 1):  # axial or coronal: horizontal axis uses x spacing.
                scale_spacing = spacing[0]
            else:  # sagittal: horizontal axis uses y spacing.
                scale_spacing = spacing[1]
            chosen_length_mm = 50  # define a 50 mm scale bar
            pixel_length = chosen_length_mm / scale_spacing

            h, w = disp_im.shape
            # placing at a defined space
            x0 = w * 0.05
            y0 = h * 0.95
            x_end = x0 + pixel_length
            ax.plot([x0, x_end], [y0, y0], "w-", linewidth=3)
            ax.text(
                (x0 + x_end) / 2,
                y0 - h * 0.03,
                f"{chosen_length_mm} mm",
                color="w",
                ha="center",
                fontsize=10,
            )

    # delete unsused axes
    for j in range(num_slices, len(axes)):
        axes[j].axis("off")

    plt.tight_layout()
    if save_filename is not None:
        try:
            plt.savefig(save_filename)
        except Exception as e:
            print(e)
    if show_plot:
        plt.show()



"""Provides classes and methods for creating irradiation geometries (stf)."""

from .generators import get_available_generators, get_generator, register_generator
from ._steeringinformation import SteeringInformation, create_stf, validate_stf
from ._beam import Beam  # , create_beam, validate_beam
from ._ray import Ray
from ._beamlet import Beamlet, IonSpot, PhotonBixel
from ._rangeshifter import RangeShifter

__all__ = [
    "get_available_generators",
    "get_generator",
    "register_generator",
    "SteeringInformation",
    "create_stf",
    "validate_stf",
    "Beam",
    "Ray",
    "Beamlet",
    "IonSpot",
    "PhotonBixel",
    "RangeShifter",
]



"""
Example of importing a patient from a MatRad file and visualizing the CT,
VOIs with matplotlib.

Also produces a simple beam configuration and visualizes the resulting radiological depth cubes by
calling the RayTracer.
"""

# Standard Library Imports
import logging
import sys

if sys.version_info < (3, 10):
    import importlib_resources as resources  # Backport for older versions
else:
    from importlib import resources  # Standard from Python 3.9+

# Third Party Imports from pyRadPlan requirements

import numpy as np
import SimpleITK as sitk
import matplotlib.pyplot as plt

# pyRadPlan Module Imports
from pyRadPlan.ct import default_hlut
from pyRadPlan.plan import IonPlan
from pyRadPlan.stf.generators import StfGeneratorIMPT
from pyRadPlan.raytracer import RayTracerSiddon
from pyRadPlan.visualization import plot_slice
from pyRadPlan.io import load_patient

# Configure the Logger to show you debug information
logging.basicConfig(level=logging.INFO)
logging.getLogger("pyRadPlan").setLevel(logging.DEBUG)

#  Read patient from provided TG119.mat file and validate data
path = resources.files("pyRadPlan.data.phantoms").joinpath("TG119.mat")
ct, cst = load_patient(path)

# Choose a slice to visualize
view_slice = int(np.round(ct.size[2] / 2))

# We store images as sitk.Image objects, but we can easily convert them to numpy arrays for
# matplotlib. Taking a "view" here means we do not perform a deep copy of the image data.
cube_hu = sitk.GetArrayViewFromImage(ct.cube_hu)
plt.imshow(cube_hu[view_slice, :, :], cmap="gray")

# Now let's visualize the VOIs from the StructureSet.
for v, voi in enumerate(cst.vois):
    mask = sitk.GetArrayViewFromImage(voi.mask)
    plt.imshow(
        (v + 1) * mask[view_slice, :, :],
        cmap="jet",
        alpha=0.5 * (mask[view_slice, :, :] > 0),
        vmin=1,
        vmax=len(cst.vois),
    )

# Since we want to do planning, we create a plan object.
# We choose protons as the radiation mode and a Generic machine included with pyRadPlan
pln = IonPlan(radiation_mode="protons", machine="Generic")

# We use the StfGeneratorIMPT to create a simple beam configuration
# with a single beam at 0 degrees
stfgen = StfGeneratorIMPT(pln)
stfgen.bixel_width = 5.0
stfgen.gantry_angles = [90.0]

# We generate the beam geometry on the CT and CST
stf = stfgen.generate(ct, cst)

# Now let's look at the water equivalent path length for our beam.
# For that, we use a default HU->rSP table to convert our CT to water-equivalent thickness and then
# call a voxel-wise RayTracing algorithm (proposed by Siddon) to calculate the radiological depth.
rt = RayTracerSiddon([ct.compute_wet(default_hlut())])
rt.debug_core_performance = True
rad_depth_cubes = rt.trace_cubes(stf[0])

# We could write the image if we want, but we don't do this by default.
# sitk.WriteImage(ct.compute_wet(default_hlut()), "wet.nrrd")
# sitk.WriteImage(rad_depth_cubes[0], "rad_depth_cubes.nrrd")

# Let's see about our range of radiological depths
rd_np = sitk.GetArrayFromImage(rad_depth_cubes[0])
rd_np.ravel()[np.isnan(rd_np.ravel())] = 0
min_rad_depth = np.min(rd_np)
max_rad_depth = np.nanmax(rd_np)
print(f"Minimum rad depth: {min_rad_depth}Maximum rad depth: {max_rad_depth}")

# Visualize
plot_slice(
    ct=ct,
    cst=cst,
    overlay=rd_np,
    view_slice=view_slice,
    plane="axial",
    overlay_unit="mm",
    overlay_rel_threshold=1e-3,
)
